import numpy as np
from pyuncertainnumber import pba
import pyuncertainnumber as pun
from pyuncertainnumber.pba.pbox_abc import Pbox


def test_precise_data():
    precise_data = np.random.normal(0, 1, 100)  # precise data case
    ub, lb = pba.KS_bounds(precise_data, alpha=0.025, display=True)
    p = pba.KS_bounds(precise_data, alpha=0.025, display=False, output_type="pbox")
    assert isinstance(p, Pbox)


def test_imprecise_data():
    precise_data = np.random.normal(0, 1, 100)  # precise data case
    ### imprecise data case
    impre_data = pba.I(lo=precise_data - 0.5, hi=precise_data + 0.5)
    u = pba.KS_bounds(impre_data, alpha=0.025, display=True, output_type="un")
    assert isinstance(u, pun.UncertainNumber)



import numpy as np
from pyuncertainnumber import area_metric
import scipy.stats as sps


def test_area_metric_on_sample_data():
    """test area_metric function on sample data and verify against scipy wasserstein_distance"""
    rng = np.random.default_rng(42)
    sample1 = rng.normal(loc=0.0, scale=1.0, size=500)
    sample2 = rng.normal(loc=0.5, scale=1.2, size=800)

    am_test = area_metric(sample1, sample2)
    am_ref = sps.wasserstein_distance(sample1, sample2)
    assert np.isclose(am_test, am_ref)



import functools


# * ---------------------shortcuts --------------------- *#
def makeUNPbox(func):

    from .pba.pbox_parametric import _bound_pcdf
    from .characterisation.uncertainNumber import UncertainNumber

    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):
        family_str = func(*args, **kwargs)
        p = _bound_pcdf(family_str, *args)
        return UncertainNumber.fromConstruct(p)

    return wrapper_decorator


def constructUN(func):
    """from a construct to create a UN"""
    from .characterisation.uncertainNumber import UncertainNumber

    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):
        p = func(*args, **kwargs)
        return UncertainNumber.fromConstruct(p)

    return wrapper_decorator


def UNtoUN(func):
    """extend constructs mapped functions to uncertain number mapped functions
    where inputs and outsuts are uncertain numbers
    """

    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):
        from .characterisation.uncertainNumber import UncertainNumber

        # transform the constructs to uncertain numbers
        uns = [c.construct for c in args]
        p = func(*uns, **kwargs)
        return UncertainNumber.fromConstruct(p)

    return wrapper_decorator


def exposeUN(func):
    """From a construct to create a UN with a choice

    example:
        >>> exposeUN(func)(*args, return_construct=True)
    """

    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):
        return_construct = kwargs.pop("return_construct", False)
        p = func(*args, **kwargs)

        if return_construct:
            return p
        from .characterisation.uncertainNumber import UncertainNumber

        return UncertainNumber.fromConstruct(p)

    return wrapper_decorator


__all__ = []


def expose_public(name, wrapper):

    def decorator(func):
        globals()[name] = wrapper(func)
        __all__.append(name)
        return func

    return decorator


# def expose_dual(public_name):
#     def decorator(func):
#         globals()[public_name] = expose_un(func)  # public wrapped version
#         return func  # private unwrapped version

#     return decorator



import numpy
import numpy.matlib as matlib
import scipy.stats as stats
from numpy.typing import NDArray


def samples_to_fuzzy_projection(ux: NDArray, c: list) -> NDArray:
    """
    Args:
        ux (NDArray): an (mxd_) array of samples, usually uniform. m is a large integer.

        c (list): a list (of length l) of subindices of coverage samples belonging to each level.
            `len(levels) < m` must yield `True`, `sum([sum(len(subi)) for subi in levels])==m` must yield `True`.

    Returns:
        fx (NDArray): returns a d-dimensional fuzzy number, i.e. an (lxdx2) array.
    """
    ux = numpy.asarray(ux, dtype=float)
    m, d_ = ux.shape
    l = len(c)
    fx = numpy.empty((l, d_, 2))
    first_empty_level = l
    for i, subi in enumerate(c):  # levels should be arranged from bottom to top
        if sum(subi) == 0:
            first_empty_level = (
                i  # set (non-empty) starting level. Yes, there may be empty levels
            )
            print(
                f"levels {i}:{l} have no sample coverage. Consider increasing the sample size."
            )
    for i in range(
        l
    ):  # levels are usually a handful, so for loop does not slow things too much
        fx[i, :, 0] = numpy.min(ux[c[i], :], axis=0)
        fx[i, :, 1] = numpy.max(ux[c[i], :], axis=0)
    if first_empty_level < l:
        for i in range(first_empty_level, l):
            fx[i, :, 0] = numpy.min(ux[c[first_empty_level - 1], :], axis=0)
            fx[i, :, 1] = numpy.max(ux[c[first_empty_level - 1], :], axis=0)
    return fx


def samples_to_fuzzy_multivariate(u: numpy.ndarray, levels: list, p: list = None):
    pass


def boxes_to_fuzzy_projection(boxes: list, p: list = None):
    """
    Args:
        boxes (list): sequence of boxes, each box is a (dx2) array. Also iterable of interval objects. Second output of the forward data-peeling algorithm.

    Returns:
        f (NDArray): an (lxdx3) fuzzy projection data structure
    """
    l = len(boxes)
    d, _ = boxes[0].shape  # (dx2)
    if p is None:
        p = numpy.arange(1 / l, 1 + 1 / l, 1 / l)  # p = 1/l
    fuzzyall = numpy.empty((l, d, 2))
    fuzzyall[:, :, :2] = numpy.asarray(
        boxes, dtype=float
    )  # f[:,:,2] = matlib.repmat(p,d,1).T
    return fuzzyall, p


def coverage_samples(lo: NDArray, hi: NDArray, m: int = 1_000):
    """
    Args:
        lo (NDArray): an (d,) array (or list) of left endpoints. Coverage means samples are generated using low-discrepancy schemes.
        hi (NDArray): an (d,) array (or list) of right endpoints, with hi > lo.

    Returns:
        u (NDArray): an (mxd_) array of coverage samples
    """
    lo, hi = numpy.asarray(lo, dtype=float), numpy.asarray(hi, dtype=float)
    if lo.size == 1:
        d = 1
    else:
        d = lo.shape[0]
    dist = stats.uniform(loc=lo, scale=hi - lo)  # <- change sampling scheme here
    return dist.rvs((m, d))


def width(x: NDArray) -> NDArray:
    """
    Args:
        x (NDArray): an interval iterable, i.e. an (dx2) array

    Returns:
        w (NDArray): the width of the intervals
    """
    x = numpy.asarray(x, dtype=float)
    return x[:, 1] - x[:, 0]



import numpy
from numpy.random import rand, randint

from .number import Interval
from .methods import intervalise

LEFT_BOUND = -1_000
RIGHT_BOUND = 1_000


def uniform_endpoints(
    n: int = 2,
    left_bound: float = None,
    right_bound: float = None,
    kind: type = float,
    shape: tuple = None,
):  # when N=2 generates two intervals
    """
    Draws endpoints from a uniform distribution.
    It was created as an interval random generator.
    """
    if left_bound is None:
        left_bound = LEFT_BOUND
    if right_bound is None:
        right_bound = RIGHT_BOUND
    if shape is None:
        if n == 1:
            if kind == float:
                improper = left_bound + rand(
                    2,
                ) * (right_bound - left_bound)
            elif kind == int:
                improper = randint(left_bound, high=right_bound, size=(2,))
            else:
                return NotImplemented
            if improper[0] <= improper[1]:
                return Interval(lo=improper[0], hi=improper[1])
            else:
                return Interval(lo=improper[1], hi=improper[0])
        else:
            if kind == float:
                improper = left_bound + rand(n, 2) * (right_bound - left_bound)
            elif kind == int:
                improper = randint(left_bound, high=right_bound, size=(n, 2))
            else:
                return NotImplemented
            improper_lo, improper_hi = improper[:, 0], improper[:, 1]
    else:
        if kind == float:
            improper_lo = left_bound + rand(*shape) * (right_bound - left_bound)
            improper_hi = left_bound + rand(*shape) * (right_bound - left_bound)
        elif kind == int:
            improper_lo = randint(left_bound, high=right_bound, size=shape)
            improper_hi = randint(left_bound, high=right_bound, size=shape)
        else:
            return NotImplemented
    swap = improper_lo >= improper_hi
    proper_lo, proper_hi = improper_lo.copy(), improper_hi.copy()
    if numpy.sum(swap) > 0:
        proper_lo[swap], proper_hi[swap] = proper_hi[swap], proper_lo[swap]
    return Interval(lo=proper_lo, hi=proper_hi)


def create_two_large_interval_matrices(shape, left_bound=0, right_bound=1):
    # shape=(100000,100000)
    x = uniform_endpoints(shape=shape, left_bound=left_bound, right_bound=right_bound)
    y = uniform_endpoints(shape=shape, left_bound=left_bound, right_bound=right_bound)
    return x, y



"""Hyperparameters for the pba"""

from dataclasses import dataclass
import numpy as np

""" hyperparameters for the pba """


@dataclass(frozen=True)  # Instances of this class are immutable.
class Params:

    steps = 200
    many = 2000

    p_lboundary = 0.001
    p_hboundary = 0.999

    # the percentiles
    p_values = np.linspace(p_lboundary, p_hboundary, steps)

    # by default
    scott_hedged_interpretation = {}

    # user-defined
    user_hedged_interpretation = {}

    # @property
    # # template for property
    # def sth(self):
    #     """ template for property"""
    #     return int(round(self.patch_window_seconds / self.stft_hop_seconds))


@dataclass(frozen=True)  # Instances of this class are immutable.
class Data:

    # scott construct p28
    skinny = [
        [1.0, 1.52],
        [2.68, 2.98],
        [7.52, 7.67],
        [7.73, 8.35],
        [9.44, 9.99],
        [3.66, 4.58],
    ]

    puffy = [
        [3.5, 6.4],
        [6.9, 8.8],
        [6.1, 8.4],
        [2.8, 6.7],
        [3.5, 9.7],
        [6.5, 9.9],
        [0.15, 3.8],
        [4.5, 4.9],
        [7.1, 7.9],
    ]

    sudret = [
        4.02,
        4.07,
        4.25,
        4.32,
        4.36,
        4.45,
        4.47,
        4.57,
        4.58,
        4.62,
        4.68,
        4.71,
        4.72,
        4.79,
        4.85,
        4.86,
        4.88,
        4.90,
        5.08,
        5.09,
        5.29,
        5.30,
        5.40,
        5.44,
        5.59,
        5.59,
        5.70,
        5.89,
        5.89,
        6.01,
    ]

    # from Scott Ioanna5.py
    k = 22
    m = 11
    n = k + m
    fdata = np.concatenate((m * [0], k * [1]))
    bdata = np.random.uniform(size=25) > 0.35
    idata = np.round(np.random.uniform(size=25) * 16)
    data = np.random.uniform(size=25) * 30
    x2 = 5 + np.random.uniform(size=25) * 30
    error = np.random.normal(size=25)

    # @property
    # # template for property
    # def sth(self):
    #     """ template for property"""
    #     return int(round(self.patch_window_seconds / self.stft_hop_seconds))


@dataclass(frozen=True)  # Instances of this class are immutable.
class Named:

    k = 22
    m = 11
    n = k + m



from functools import singledispatch
import numpy as np
from . import intervalise, Interval
from ...nlp.language_parsing import parse_interval_expression, hedge_interpret
from numbers import Number

""" operations for generic Interval objects """

# see the hedged interpretation for Interval in `nlp/language_parsing.py`


def parse_bounds(b):
    """top-level function that universally parses scalar and vector bounds"""
    try:
        return wc_scalar_interval(b)
    except Exception:
        return make_vec_interval(b)


# * ---------------------make scalar interval object --------------------- *#


def wc_scalar_interval_feature(*args):
    """wildcard scalar interval

        This function is used to parse a scalar bound into an Interval object.
        It can handle various input types such as lists, tuples, and strings.
        If the input is a string, it attempts to interpret it using the
        `hedge_interpret` function or parse it as an interval expression.
        If the input is a single number, it creates an Interval with that number
        as both bounds.


    note:
        This function is a beta version of `wc_scalar_interval` which is meant to test the API signature.
        If run into error, then resort back to `wc_scalar_interval`.
    """
    from ...characterisation.uncertainNumber import UncertainNumber as UN

    if len(args) == 1:
        bound = args[0]
    elif len(args) == 2:
        return Interval(*args)
    else:
        raise ValueError("wc_scalar_interval only accepts 1 or 2 arguments")

    if isinstance(bound, list):
        return Interval(*bound)
    elif isinstance(bound, tuple):
        return Interval(*bound)
    elif isinstance(bound, Interval):
        return bound
    elif isinstance(bound, Number):
        return Interval(bound, bound)
    elif isinstance(bound, str):
        try:
            return hedge_interpret(bound)
        except Exception:
            pass
        try:
            return parse_interval_expression(bound)
        except Exception:
            raise ValueError("Invalid input")
    elif isinstance(bound, UN):
        return bound.construct
    else:
        raise TypeError("Unsupported type for interval creation")


def wc_scalar_interval(bound):
    """wildcard scalar interval

    This function is used to parse a scalar bound into an Interval object.
    It can handle various input types such as lists, tuples, and strings.
    If the input is a string, it attempts to interpret it using the
    `hedge_interpret` function or parse it as an interval expression.
    If the input is a single number, it creates an Interval with that number
    as both bounds.
    """
    from ...characterisation.uncertainNumber import UncertainNumber as UN

    if isinstance(bound, list):
        return Interval(*bound)
    elif isinstance(bound, tuple):
        return Interval(*bound)
    elif isinstance(bound, Interval):
        return bound
    elif isinstance(bound, Number):
        return Interval(bound, bound)
    elif isinstance(bound, str):
        try:
            return hedge_interpret(bound)
        except Exception:
            pass
        try:
            return parse_interval_expression(bound)
        except Exception:
            raise ValueError("Invalid input")
    elif isinstance(bound, UN):
        return bound.construct
    else:
        raise TypeError("Unsupported type for interval creation")


""" old implementation

note: it is deprecated due to it cannot handle custom Class e.g. UN 
because it has to import UN which creates circular import issue.
"""


# @singledispatch
# def wc_scalar_interval(bound):
#     """wildcard scalar interval"""
#     return Interval(bound)


# @wc_scalar_interval.register(list)
# def _list(bound: list):
#     return Interval(*bound)


# @wc_scalar_interval.register(tuple)
# def _tuple(bound: tuple):
#     return Interval(*bound)


# @wc_scalar_interval.register(Interval)
# def _marco_interval_like(bound: Interval):
#     return bound


# @wc_scalar_interval.register(Number)
# def _scalar(bound: Number):
#     return Interval(bound, bound)


# @wc_scalar_interval.register(str)
# def _scalar(bound: str):

#     try:
#         return hedge_interpret(bound)
#     except Exception:
#         pass

#     try:
#         return parse_interval_expression(bound)
#     except Exception:
#         raise ValueError("Invalid input")


# * ---------------------make vector interval object --------------------- *#


def make_vec_interval(vec):
    """parse an array-like structure into a vector interval

    For most part, it works same to `intervalise`, except that this function
    can also handle a list of UN objects.

    example:
        >>> a, b = pba.I(1, 2), pba.I(3, 4)
        >>> make_vec_interval([a, b])
        Interval([1, 3], [2, 4])

    """
    from ...characterisation.uncertainNumber import UncertainNumber as UN

    assert len(vec) > 1, "Interval must have more than one element"

    if isinstance(vec, Interval):
        return vec
    # elif isinstance(vec, list):
    #     if all(isinstance(v, Interval) for v in vec):
    #         return Interval(lo=[v.lo for v in vec], hi=[v.hi for v in vec])
    elif isinstance(vec[0], Interval | list | tuple | np.ndarray):
        try:
            return intervalise(vec)
        except Exception:
            return intervalise(list(vec))
    elif isinstance(vec[0], UN):
        return intervalise([v.construct for v in vec])
    else:
        raise NotImplementedError


# * ---------------------mean func --------------------- *#


@singledispatch
def mean(x):
    return np.mean(x)


@mean.register(np.ndarray)
def _arraylike(x):
    return np.mean(x)


@mean.register(Interval)
def _intervallike(x):
    return sum(x) / len(x)



from __future__ import annotations
from typing import TYPE_CHECKING
from scipy.interpolate import interp1d
import numpy as np
from .pbox_abc import Staircase, Pbox
from .params import Params
from ..characterisation.utils import tranform_ecdf


if TYPE_CHECKING:
    from .utils import eCDF_bundle


# # TODO to integrate it
# def pbox_fromeF(a: eCDF_bundle, b: eCDF_bundle):
#     """pbox from emipirical CDF bundle
#     args:
#         - a : CDF bundle of lower extreme F;
#         - b : CDF bundle of upper extreme F;
#     """
#     from .pbox_abc import Staircase

#     # TODO currently the interpolation is not perfect
#     p_lo, q_lo = interpolate_p(a.probabilities, a.quantiles)
#     p_hi, q_hi = interpolate_p(b.probabilities, b.quantiles)
#     return Staircase(left=q_lo, right=q_hi)


""" initially used for cbox next-value distribution """


def pbox_from_pseudosamples(samples):
    """a tmp constructor for pbox/cbox from approximate solution of the confidence/next value distribution

    args:
        samples (nd.array): the approximate Monte Carlo samples of the confidence/next value distribution

    note:
        ecdf is estimted from the samples and bridge to pbox/cbox
    """
    return Pbox(tranform_ecdf(samples, display=False))


def interpolate_p(p, q):
    """interpolate the cdf bundle for discrete distribution or ds structure

    note:
        - p: probabilities (x)
        - q: quantiles (y)

    return: a tuple of new (p, q)
    """

    f = interp1d(p, q, kind="next", fill_value=(p[0], p[-1]), bounds_error=False)
    # range
    # new_p = np.linspace(p[0], p[-1], Params.steps)
    new_p = Params.p_values
    new_q = f(new_p)
    return new_p, new_q



from typing import Callable
from rich.progress import track
import numpy as np
from SALib.sample import sobol as sobol_sample
from SALib.analyze import sobol as sobol_analyze
from matplotlib import pyplot as plt


def sobol_analysis(
    x: list, f: Callable, output_names=None, plot_results: bool = True, **kwargs
):
    """Performs a Sobol sensitivity analysis using the SALibe library.

    Args:
        x (list): A list of ``UncertainNumber`` objects defining the input variables.
        f (Callable): The model function to be analyzed.
        output_names (list, optional): A list of strings to name the model outputs.
                                       Defaults to None.
        plot_results (bool, optional): If True, generates and displays plots of the
                                       results. Defaults to True.
        **kwargs: Additional keyword arguments to pass to SALib's ``sample`` and
                  ``analyze`` functions. For example, ``N`` (the number of samples)
                  must be a power of 2 (e.g., 1024, 2048) and defaults to 1024.

    Returns:
        dict: A dictionary where keys are the output names and values are the
              corresponding SALib analysis result objects.

    .. tip::
        This function is designed to work with a list of ``UncertainNumber`` objects
        that are defined as distributions. It evaluates the sensitivity of a model
        function ``f``, which can have single or multiple outputs, to these inputs.

        The code performs the following steps:
            1.  **Problem Definition**: A SALib `problem` dictionary is constructed from the
                input list `X` of UncertainNumber objects. This dictionary defines the
                variables, their names, distributions, and parameters.

            2.  **Vectorization Auto-Detection**: The function automatically probes the
                model function `f` to determine if it is "vectorized" (i.e., can
                process all samples at once with NumPy) or "non-vectorized" (processes
                one sample at a time). This allows for flexible use without needing to
                modify the model function.

            3.  **Direct Sampling**: Input samples are generated using `SALib.sample.sobol.sample`.
                This function is passed the problem definition directly.
                **Important Note**: When using this direct method, SALib performs a
                linear scaling of the Sobol sequence from the [0, 1] domain to the
                `bounds` provided. It does *not* perform a distributional transformation
                (e.g., using the inverse CDF). For a 'gaussian' distribution defined
                with bounds `[mean, std_dev]`, the generated samples will be uniformly
                distributed between `mean` and `std_dev`, not normally distributed.

            4.  **Model Evaluation**: The model `f` is evaluated using the generated samples.
                The evaluation is performed efficiently if the model is vectorized, or
                iteratively (with a progress bar) if it is not.

            5.  **Output Handling**: The function gracefully handles models that return
                single or multiple outputs, normalizing them for analysis.

            6.  **Sobol Analysis**: For each model output, `SALib.analyze.sobol.analyze`
                is called to compute the first-order (S1), second-order (S2), and
                total-order (ST) sensitivity indices.

            7.  **Plotting**: If `plot_results` is True, bar charts of the S1 and ST
                indices are generated for each output.

    Example:
        >>> from pyuncertain import UncertainNumber as UN
        >>>
        >>> # 1. Define uncertain inputs
        >>> inputs = [
        ...     UN(name='x1', essence='distribution', distribution_parameters=["uniform", (0, 1)]),
        ...     UN(name='x2', essence='distribution', distribution_parameters=["gaussian", (0.5, 0.1)])
        ... ]
        >>>
        >>> # 2. Define a model function (non-vectorized)
        >>> def my_model(x):
        ...     return x[0]**2 + x[1]
        >>>
        >>> # 3. Run the analysis
        >>> results = sobol_analysis(inputs, my_model, print_to_console = True, plot_results: bool = True, N=512)
        >>>
        >>> # 4. Print results for the first output
        >>> results['Output_1']
        >>>

    """
    try:
        # === STAGE 1: Build the problem dictionary ===
        dist_map = {"gaussian": "norm", "uniform": "unif", "triangular": "triang"}
        names, bounds, dists = [], [], []

        for un in x:
            var_name = un.name
            dist_name, dist_params = un.distribution_parameters
            if dist_name not in dist_map:
                raise ValueError(f"Distribution '{dist_name}' is not supported.")
            names.append(var_name)
            bounds.append(list(dist_params))
            dists.append(dist_map[dist_name])

        problem = {"num_vars": len(x), "names": names, "bounds": bounds, "dists": dists}

        # === STAGE 2: Auto-Detect Vectorization ===
        is_vectorized = False
        try:
            probe_input = np.zeros((2, problem["num_vars"]))
            f(probe_input)
            print(f(probe_input))
            is_vectorized = True
            print("INFO: Auto-detected vectorized model.")
        except Exception:
            print(
                "INFO: Auto-detected non-vectorized model. Using sample-by-sample evaluation."
            )

        # === STAGE 3: Sample and Evaluate Model ===
        KNOWN_SAMPLING_ARGS = [
            "N",
            "calc_second_order",
            "scramble",
            "skip_values",
            "seed",
        ]
        sample_args, analyze_args = {}, {}
        for key, value in kwargs.items():
            if key in KNOWN_SAMPLING_ARGS:
                sample_args[key] = value
            else:
                analyze_args[key] = value

        sample_args.setdefault("N", 1024)
        sample_args.setdefault("calc_second_order", True)
        analyze_args.setdefault("print_to_console", True)

        param_values = sobol_sample.sample(problem, **sample_args)

        if is_vectorized:
            Y = f(param_values)
        else:
            Y = [
                f(row)
                for row in track(param_values, description="Processing samples...")
            ]
            print("Y", Y)

        # === STAGE 4: Normalize Output and Analyze ===
        if isinstance(Y, (list, tuple)) and not np.isscalar(Y[0]):
            Y_multi_output = (
                np.array(Y) if isinstance(Y[0], (list, tuple)) else np.column_stack(Y)
            )
        elif isinstance(Y, list):
            Y_multi_output = np.array(Y).reshape(-1, 1)
        elif Y.ndim == 1:
            Y_multi_output = Y.reshape(-1, 1)
        elif Y.ndim == 2:
            Y_multi_output = Y
        else:
            raise ValueError(f"Model output has an unsupported shape: {Y.shape}")

        num_outputs = Y_multi_output.shape[1]
        if output_names is None:
            output_names = [f"Output_{i+1}" for i in range(num_outputs)]
        elif len(output_names) != num_outputs:
            raise ValueError(
                f"Provided {len(output_names)} names, but model has {num_outputs} outputs."
            )

        all_results = {}
        for i, name in enumerate(output_names):
            print(f"\nResults for '{name}':")
            Si = sobol_analyze.analyze(problem, Y_multi_output[:, i], **analyze_args)
            all_results[name] = Si

            # all_results[name]
            if plot_results:

                all_results[name].plot()
                plt.title(f"Sobol Indices for: {name}")
                plt.xlabel("Parameters")
                plt.show()

        return all_results

    except Exception as e:
        print(f"\n>>> SCRIPT FAILED <<<")
        import traceback

        traceback.print_exc()
        return None



import numpy as np
import scipy.stats as sps
from pyuncertainnumber import make_vec_interval, Interval


def cauchy_deviate_method(
    func,
    input_vector_interval: list[Interval],
    n_sam: int = 200,
) -> Interval:
    """Cauchy Deviate Method for interval propagation

    args:
        func: The function, vectorised style, to be evaluated.
        input_vector_interval: The input vector of intervals.
        n_sam: The number of samples to draw from each Cauchy distribution.

    note:
        The function must be vectorised, i.e. it must be able to take in a
        2D array of shape (n, d) and return a 1D array of shape (n,).
    """
    # enforce vec-Interval
    input_vector_interval = make_vec_interval(input_vector_interval)

    # of dimension d
    nominal_measurement = input_vector_interval.mid

    # of dimension d
    scalar_param = input_vector_interval.rad
    mean_y, y_scalar_param = cauchy_deviate_raw(
        func, nominal_measurement, scalar_param, n_sam
    )
    return Interval.from_meanform(mean_y, y_scalar_param)


def cauchy_deviate_raw(
    func, nominal_measurement: np.ndarray, scalar_param: np.ndarray, n_sam: int = 200
) -> tuple:
    """Raw implementation of the Cauchy Deviate Method

    Args:
        func: The function, vectorised style, to be evaluated.
        nominal_measurement: The nominal measurement values, i.e. x tilda.
        scalar_param: The scalar parameters for each component Cauchy distribution, i.e. Delta.
        n_sam: The number of samples to draw from each Cauchy distribution.

    Returns:
        mean_y and Delta_y
    """

    # of dimension d
    nominal_measurement = np.array(nominal_measurement)

    # of dimension d
    scalar_param = np.array(scalar_param)
    samples = sps.cauchy.rvs(
        scale=scalar_param[:, None], size=(len(scalar_param), n_sam)
    ).T  # samples of shape (n_sam, d)

    # find the max of each row
    K = np.max(samples, axis=1)

    normalised_ = samples / K[:, None]
    semi_actual_inputs = nominal_measurement - normalised_

    # vec
    f_mid = func(nominal_measurement[np.newaxis, :])
    f_semi = func(semi_actual_inputs)
    y_diffs = f_mid - f_semi

    # normalised output difference
    normalised_y_diff = y_diffs * K

    outout_scalar_param = sps.cauchy.fit(normalised_y_diff, floc=0)

    return f_mid.item(), outout_scalar_param[1].item()



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyuncertainnumber import Interval

from .bo import BayesOpt
from .ga import GA
import numpy as np


def get_range_BO(
    f: callable,
    design_bounds: list | np.ndarray,
    acquisition_function: str = "UCB",
    verbose: bool = False,
    **kwargs,
) -> tuple[Interval, dict]:
    """Compute the range of a black-box function using BayesOpt with vectorised or iterable function signature

    args:
        f (callable): function as the objective to be optimized, with vectorised or iterable function signature

        design_bounds (list | np.ndarray): nested list or 2D array, containing the lower and upper bounds for each dimension; e.g. [[0, 1], [0, 1]] for 2D input space.

        acquisition_function (str or callable, optional): the acquisition function to be used, e.g. 'UCB', 'EI', 'PI'. If None, defaults to 'UCB'.

        verbose (Boolean): if True, prints the optimization progress

        **kwargs: additional keyword arguments for the BayesOpt class. For example, one can pass 'num_explorations', 'num_iterations', etc.

    tip:
        for a less verbose output, use (convergence_curve=False, progress_bar=False)

    return:
        Tuple[Interval, dict]: A tuple containing:
            - response_itvl: The interval of the minimum and maximum from the optimization of the black-box function.
            - opt_hint: A dictionary with optimal input points for the minimum and maximum values.

    """
    from ..pba.intervals.number import Interval

    if not verbose:  # quiet mode
        V = 0
    else:
        V = 1

    min_task = BayesOpt(
        f=f,
        task="minimisation",
        design_bounds=design_bounds,
        acquisition_function=acquisition_function,
        **kwargs,
    )
    min_task.run(verbose=V)
    min_target = min_task.optimal_target

    max_task = BayesOpt(
        f=f,
        task="maximisation",
        design_bounds=design_bounds,
        acquisition_function=acquisition_function,
        **kwargs,
    )

    max_task.run(verbose=V)
    max_target = max_task.optimal_target

    # return 1: the interval of min and max
    response_itvl = Interval(min_target, max_target)

    # return 2: the mapping associated with the optimisation
    opt_hint = {
        "min": min_task.optimal,
        "max": max_task.optimal,
    }
    return response_itvl, opt_hint


def get_range_BO_raw(
    f: callable,
    design_bounds: list,
    acquisition_function="UCB",
    verbose=False,
    **kwargs,
):
    """Compute the range of a black-box function using BayesOpt with arguments-signature function;

    args:
        f: callable function to be optimized

        dimension (int): the number of dimensions of the input space

        design_bounds (list): each tuple contains the lower and upper bounds for each dimension

        acquisition_function (str or callable, optional): the acquisition function to be used, e.g. 'UCB', 'EI', 'PI'. If None, defaults to 'UCB'.

        verbose (Boolean): if True, prints the optimization progress

        **kwargs: additional keyword arguments for the BayesOpt class

    tip:
        for a less verbose output, use (convergence_curve=False, progress_bar=False)

    return:
        Tuple[Interval, dict]: A tuple containing:
            - response_itvl: The interval of the minimum and maximum from the optimization of the black-box function.
            - opt_hint: A dictionary with optimal input points for the minimum and maximum values.

    """
    from ..pba.intervals.number import Interval

    if not verbose:  # quiet mode
        V = 0
    else:
        V = 1

    min_task = BayesOpt(
        f=f,
        task="minimisation",
        design_bounds=design_bounds,
        acquisition_function=acquisition_function,
        **kwargs,
    )
    min_task.run(verbose=V)
    min_target = min_task.optimal_target

    max_task = BayesOpt(
        f=f,
        task="maximisation",
        design_bounds=design_bounds,
        acquisition_function=acquisition_function,
        **kwargs,
    )

    max_task.run(verbose=V)
    max_target = max_task.optimal_target

    # return 1: the interval of min and max
    response_itvl = Interval(min_target, max_target)

    # return 2: the mapping associated with the optimisation
    opt_hint = {
        "min": min_task.optimal,
        "max": max_task.optimal,
    }
    return response_itvl, opt_hint


def get_range_GA(
    f: callable, dimension: int, varbound, algorithm_param=None, verbose=False, **kwargs
):
    """compute the range of the black-box function using GA

    args:
        varbound (np.ndarray): The variable bounds for the optimization.

    return:
        Tuple[Interval, dict]: A tuple containing:
            - response_itvl: The interval of the minimum and maximum from the optimization of the black-box function.
            - opt_hint: A dictionary with optimal input points for the minimum and maximum values.

    note:
        It's suggested to use `EpistemicDomain` which facilitates the specification of varbound.
    """

    from ..pba.intervals.number import Interval

    if not verbose:
        kwargs["convergence_curve"] = False
        kwargs["progress_bar"] = False

    min_task = GA(f, task="minimisation", dimension=dimension, varbound=varbound)
    min_task.run(algorithm_param=algorithm_param, **kwargs)
    min_target = min_task.optimal_target

    max_task = GA(f, task="maximisation", dimension=dimension, varbound=varbound)
    max_task.run(algorithm_param=algorithm_param, **kwargs)
    max_target = max_task.optimal_target

    # return 1: the interval of min and max
    response_itvl = Interval(min_target, max_target)

    # return 2: the mapping associated with the optimisation
    opt_hint = {
        "min": min_task.optimal,
        "max": max_task.optimal,
    }

    return response_itvl, opt_hint



import numpy as np
import operator
import pytest
from pyuncertainnumber import pba


@pytest.fixture
def construct_dss():
    d1 = pba.DSS(intervals=[[1, 5], [3, 6]], masses=[0.5, 0.5])
    d2 = pba.DSS(intervals=[[2, 3], [2, 9]], masses=[0.5, 0.5])
    return d1, d2


# Map operation names to Python operator functions
OPS = {
    "add": operator.add,
    "sub": operator.sub,
    "mul": operator.mul,
    "truediv": operator.truediv,
}


UNARY = {
    "log": np.log,
    "exp": np.exp,
    "sqrt": np.sqrt,
}


@pytest.mark.parametrize("op_name", list(OPS.keys()), ids=list(OPS.keys()))
def test_arithmetic_operations(construct_dss, op_name):
    d1, d2 = construct_dss
    op = OPS[op_name]

    # keep your logic; just avoid recomputing to_pbox()
    p1 = d1.to_pbox()
    p2 = d2.to_pbox()

    c = op(p1, p2)

    # Your own style of check
    assert c == op(p1, p2)


@pytest.mark.parametrize("op_name", list(UNARY.keys()), ids=list(UNARY.keys()))
def test_unary_operations(construct_dss, op_name):
    d1, _ = construct_dss
    op = UNARY[op_name]

    p = d1.to_pbox()
    c = op(p)

    # Keep your own style of check
    assert c == op(p)



import pytest
from pyuncertainnumber import pba


# * ----------------------- mixture operation


def test_mix_dss():

    d1 = pba.DempsterShafer(intervals=[[1, 5], [3, 6]], masses=[0.5, 0.5])
    d2 = pba.DempsterShafer(intervals=[[2, 3], [2, 9]], masses=[0.5, 0.5])

    dd = pba.stochastic_mixture(d1, d2)


def test_mix_intervals():
    s = pba.stacking([[1, 3], [2, 4]], weights=[0.5, 0.5], display=True)
    new_dss = pba.DSS(intervals=[[1, 3], [2, 4]], masses=[0.5, 0.5])
    assert s == new_dss


# * ----------------------- envelope operation


def test_env_distribution():
    a = pba.normal(3, 1)
    b = pba.uniform(5, 8)
    c = pba.normal(13, 2)
    t = pba.envelope(a, b, c)
    assert a in t
    assert b in t
    assert c in t


def test_env_pbox():

    a = pba.normal([3, 5], 1)
    b = pba.uniform([3, 5], [6, 9])
    t = pba.envelope(a, b)

    assert a in t
    assert b in t


def test_env_intervals():
    a = pba.I(2, 6)
    b = pba.I(-2, 9)
    t = pba.envelope(a, b)

    assert isinstance(t, pba.Interval)
    # assert a in t
    # assert b in t


# * ----------------------- imposition operation


def test_imp_pbox():
    a = pba.normal([3, 7], 1)
    b = pba.uniform([3, 5], [6, 9])
    i = pba.imposition(a, b)
    assert i in a
    assert i in b



"""
--------------------------
Created Feb 2022
Marco De Angelis
github.com/marcodeangelis
MIT License
--------------------------
"""

import numpy


def multiply(s, o):
    s_lo, s_hi, o_lo, o_hi = s.lo, s.hi, o.lo, o.hi
    if s.scalar & o.scalar:
        if (s_lo >= 0) & (o_lo >= 0):  # A+ B+
            l, h = s_lo * o_lo, s_hi * o_hi
        if (s_lo >= 0) & ((o_lo < 0) & (o_hi > 0)):  # A+ B0
            l, h = s_hi * o_lo, s_hi * o_hi
        if (s_lo >= 0) & (o_hi <= 0):  # A+ B-
            l, h = s_hi * o_lo, s_lo * o_hi
        if ((s_lo < 0) & (s_hi > 0)) & (o_lo >= 0):  # A0 B+
            l, h = s_lo * o_hi, s_hi * o_hi
        if ((s_lo < 0) & (s_hi > 0)) & ((o_lo < 0) & (o_hi > 0)):  # A0 B0
            l = numpy.min((s_lo * o_hi, s_hi * o_lo, s_lo * o_lo, s_hi * o_hi), axis=0)
            h = numpy.max((s_lo * o_lo, s_hi * o_hi, s_lo * o_hi, s_hi * o_lo), axis=0)
        if ((s_lo < 0) & (s_hi > 0)) & (o_hi <= 0):  # A0 B-
            l, h = s_hi * o_lo, s_lo * o_lo
        if (s_hi <= 0) & (o_lo >= 0):  # A- B+
            l, h = s_lo * o_hi, s_hi * o_lo
        if (s_hi <= 0) & ((o_lo < 0) & (o_hi > 0)):  # A- B0
            l, h = s_lo * o_hi, s_lo * o_lo
        if (s_hi <= 0) & (o_hi <= 0):  # A- B-
            l, h = s_hi * o_hi, s_lo * o_lo
    elif s_lo.shape == o_lo.shape:
        l, h = numpy.empty(s_lo.shape), numpy.empty(s_lo.shape)
        pp = (s_lo >= 0) & (o_lo >= 0)  # A+ B+
        l[pp] = s_lo[pp] * o_lo[pp]
        h[pp] = s_hi[pp] * o_hi[pp]
        pz = (s_lo >= 0) & ((o_lo < 0) & (o_hi > 0))  # A+ B0
        l[pz] = s_hi[pz] * o_lo[pz]
        h[pz] = s_hi[pz] * o_hi[pz]
        pn = (s_lo >= 0) & (o_hi <= 0)  # A+ B-
        l[pn] = s_hi[pn] * o_lo[pn]
        h[pn] = s_lo[pn] * o_hi[pn]
        zp = ((s_lo < 0) & (s_hi > 0)) & (o_lo >= 0)  # A0 B+
        l[zp] = s_lo[zp] * o_hi[zp]
        h[zp] = s_hi[zp] * o_hi[zp]
        zz = ((s_lo < 0) & (s_hi > 0)) & ((o_lo < 0) & (o_hi > 0))  # A0 B0
        l[zz] = numpy.min(
            (
                s_lo[zz] * o_hi[zz],
                s_hi[zz] * o_lo[zz],
                s_lo[zz] * o_lo[zz],
                s_hi[zz] * o_hi[zz],
            ),
            axis=0,
        )
        h[zz] = numpy.max(
            (
                s_lo[zz] * o_lo[zz],
                s_hi[zz] * o_hi[zz],
                s_lo[zz] * o_hi[zz],
                s_hi[zz] * o_lo[zz],
            ),
            axis=0,
        )
        zn = ((s_lo < 0) & (s_hi > 0)) & (o_hi <= 0)  # A0 B-
        l[zn] = s_hi[zn] * o_lo[zn]
        h[zn] = s_lo[zn] * o_lo[zn]
        np = (s_hi <= 0) & (o_lo >= 0)  # A- B+
        l[np] = s_lo[np] * o_hi[np]
        h[np] = s_hi[np] * o_lo[np]
        nz = (s_hi <= 0) & ((o_lo < 0) & (o_hi > 0))  # A- B0
        l[nz] = s_lo[nz] * o_hi[nz]
        h[nz] = s_lo[nz] * o_lo[nz]
        nn = (s_hi <= 0) & (o_hi <= 0)  # A- B-
        l[nn] = s_hi[nn] * o_hi[nn]
        h[nn] = s_lo[nn] * o_lo[nn]
    elif s.scalar:
        l, h = numpy.empty(o_lo.shape), numpy.empty(o_lo.shape)
        pp = (s_lo >= 0) & (o_lo >= 0)  # A+ B+
        l[pp] = s_lo * o_lo[pp]
        h[pp] = s_hi * o_hi[pp]
        pz = (s_lo >= 0) & ((o_lo < 0) & (o_hi > 0))  # A+ B0
        l[pz] = s_hi * o_lo[pz]
        h[pz] = s_hi * o_hi[pz]
        pn = (s_lo >= 0) & (o_hi <= 0)  # A+ B-
        l[pn] = s_hi * o_lo[pn]
        h[pn] = s_lo * o_hi[pn]
        zp = ((s_lo < 0) & (s_hi > 0)) & (o_lo >= 0)  # A0 B+
        l[zp] = s_lo * o_hi[zp]
        h[zp] = s_hi * o_hi[zp]
        zz = ((s_lo < 0) & (s_hi > 0)) & ((o_lo < 0) & (o_hi > 0))  # A0 B0
        l[zz] = numpy.min(
            (s_lo * o_hi[zz], s_hi * o_lo[zz], s_lo * o_lo[zz], s_hi * o_hi[zz]), axis=0
        )
        h[zz] = numpy.max(
            (s_lo * o_lo[zz], s_hi * o_hi[zz], s_lo * o_hi[zz], s_hi * o_lo[zz]), axis=0
        )
        zn = ((s_lo < 0) & (s_hi > 0)) & (o_hi <= 0)  # A0 B-
        l[zn] = s_hi * o_lo[zn]
        h[zn] = s_lo * o_lo[zn]
        np = (s_hi <= 0) & (o_lo >= 0)  # A- B+
        l[np] = s_lo * o_hi[np]
        h[np] = s_hi * o_lo[np]
        nz = (s_hi <= 0) & ((o_lo < 0) & (o_hi > 0))  # A- B0
        l[nz] = s_lo * o_hi[nz]
        h[nz] = s_lo * o_lo[nz]
        nn = (s_hi <= 0) & (o_hi <= 0)  # A- B-
        l[nn] = s_hi * o_hi[nn]
        h[nn] = s_lo * o_lo[nn]
    elif o.scalar:
        l, h = numpy.empty(s_lo.shape), numpy.empty(s_lo.shape)
        pp = (s_lo >= 0) & (o_lo >= 0)  # A+ B+
        l[pp] = s_lo[pp] * o_lo
        h[pp] = s_hi[pp] * o_hi
        pz = (s_lo >= 0) & ((o_lo < 0) & (o_hi > 0))  # A+ B0
        l[pz] = s_hi[pz] * o_lo
        h[pz] = s_hi[pz] * o_hi
        pn = (s_lo >= 0) & (o_hi <= 0)  # A+ B-
        l[pn] = s_hi[pn] * o_lo
        h[pn] = s_lo[pn] * o_hi
        zp = ((s_lo < 0) & (s_hi > 0)) & (o_lo >= 0)  # A0 B+
        l[zp] = s_lo[zp] * o_hi
        h[zp] = s_hi[zp] * o_hi
        zz = ((s_lo < 0) & (s_hi > 0)) & ((o_lo < 0) & (o_hi > 0))  # A0 B0
        l[zz] = numpy.min(
            (s_lo[zz] * o_hi, s_hi[zz] * o_lo, s_lo[zz] * o_lo, s_hi[zz] * o_hi), axis=0
        )
        h[zz] = numpy.max(
            (s_lo[zz] * o_lo, s_hi[zz] * o_hi, s_lo[zz] * o_hi, s_hi[zz] * o_lo), axis=0
        )
        zn = ((s_lo < 0) & (s_hi > 0)) & (o_hi <= 0)  # A0 B-
        l[zn] = s_hi[zn] * o_lo
        h[zn] = s_lo[zn] * o_lo
        np = (s_hi <= 0) & (o_lo >= 0)  # A- B+
        l[np] = s_lo[np] * o_hi
        h[np] = s_hi[np] * o_lo
        nz = (s_hi <= 0) & ((o_lo < 0) & (o_hi > 0))  # A- B0
        l[nz] = s_lo[nz] * o_hi
        h[nz] = s_lo[nz] * o_lo
        nn = (s_hi <= 0) & (o_hi <= 0)  # A- B-
        l[nn] = s_hi[nn] * o_hi
        h[nn] = s_lo[nn] * o_lo
    return l, h


def divide(s, o):
    s_lo, s_hi, o_lo, o_hi = s.lo, s.hi, o.lo, o.hi
    other_straddle_zero = numpy.any((o_lo.flatten() <= 0) & (o_hi.flatten() >= 0))
    if other_straddle_zero:
        raise ZeroDivisionError
    if s.scalar & o.scalar:
        if (s_lo >= 0) & (o_lo > 0):  # A+ B+
            l, h = s_lo / o_hi, s_hi / o_lo
        if ((s_lo < 0) & (s_hi > 0)) & (o_lo > 0):  # A0 B+
            l, h = s_lo / o_lo, s_hi / o_lo
        if (s_hi <= 0) & (o_lo >= 0):  # A- B+
            l, h = s_lo / o_lo, s_hi / o_hi
        if (s_lo >= 0) & (o_hi <= 0):  # A+ B-
            l, h = s_hi / o_hi, s_lo / o_lo
        if ((s_lo < 0) & (s_hi > 0)) & (o_hi <= 0):  # A0 B-
            l, h = s_hi / o_hi, s_lo / o_hi
        if (s_hi <= 0) & (o_hi <= 0):  # A- B-
            l, h = s_hi / o_lo, s_lo / o_hi
    elif s_lo.shape == o_lo.shape:
        l, h = numpy.empty(s_lo.shape), numpy.empty(s_lo.shape)
        pp = (s_lo >= 0) & (o_lo > 0)  # A+ B+
        l[pp] = s_lo[pp] / o_hi[pp]
        h[pp] = s_hi[pp] / o_lo[pp]
        zp = ((s_lo < 0) & (s_hi > 0)) & (o_lo > 0)  # A0 B+
        l[zp] = s_lo[zp] / o_lo[zp]
        h[zp] = s_hi[zp] / o_lo[zp]
        np = (s_hi <= 0) & (o_lo >= 0)  # A- B+
        l[np] = s_lo[np] / o_lo[np]
        h[np] = s_hi[np] / o_hi[np]
        pn = (s_lo >= 0) & (o_hi <= 0)  # A+ B-
        l[pn] = s_hi[pn] / o_hi[pn]
        h[pn] = s_lo[pn] / o_lo[pn]
        zn = ((s_lo < 0) & (s_hi > 0)) & (o_hi <= 0)  # A0 B-
        l[zn] = s_hi[zn] / o_hi[zn]
        h[zn] = s_lo[zn] / o_hi[zn]
        nn = (s_hi <= 0) & (o_hi <= 0)  # A- B-
        l[nn] = s_hi[nn] / o_lo[nn]
        h[nn] = s_lo[nn] / o_hi[nn]
    elif s.scalar:
        l, h = numpy.empty(o_lo.shape), numpy.empty(o_lo.shape)
        pp = (s_lo >= 0) & (o_lo > 0)  # A+ B+
        l[pp] = s_lo / o_hi[pp]
        h[pp] = s_hi / o_lo[pp]
        zp = ((s_lo < 0) & (s_hi > 0)) & (o_lo > 0)  # A0 B+
        l[zp] = s_lo / o_lo[zp]
        h[zp] = s_hi / o_lo[zp]
        np = (s_hi <= 0) & (o_lo >= 0)  # A- B+
        l[np] = s_lo / o_lo[np]
        h[np] = s_hi / o_hi[np]
        pn = (s_lo >= 0) & (o_hi <= 0)  # A+ B-
        l[pn] = s_hi / o_hi[pn]
        h[pn] = s_lo / o_lo[pn]
        zn = ((s_lo < 0) & (s_hi > 0)) & (o_hi <= 0)  # A0 B-
        l[zn] = s_hi / o_hi[zn]
        h[zn] = s_lo / o_hi[zn]
        nn = (s_hi <= 0) & (o_hi <= 0)  # A- B-
        l[nn] = s_hi / o_lo[nn]
        h[nn] = s_lo / o_hi[nn]
    elif o.scalar:
        l, h = numpy.empty(s_lo.shape), numpy.empty(s_lo.shape)
        pp = (s_lo >= 0) & (o_lo > 0)  # A+ B+
        l[pp] = s_lo[pp] / o_hi
        h[pp] = s_hi[pp] / o_lo
        zp = ((s_lo < 0) & (s_hi > 0)) & (o_lo > 0)  # A0 B+
        l[zp] = s_lo[zp] / o_lo
        h[zp] = s_hi[zp] / o_lo
        np = (s_hi <= 0) & (o_lo >= 0)  # A- B+
        l[np] = s_lo[np] / o_lo
        h[np] = s_hi[np] / o_hi
        pn = (s_lo >= 0) & (o_hi <= 0)  # A+ B-
        l[pn] = s_hi[pn] / o_hi
        h[pn] = s_lo[pn] / o_lo
        zn = ((s_lo < 0) & (s_hi > 0)) & (o_hi <= 0)  # A0 B-
        l[zn] = s_hi[zn] / o_hi
        h[zn] = s_lo[zn] / o_hi
        nn = (s_hi <= 0) & (o_hi <= 0)  # A- B-
        l[nn] = s_hi[nn] / o_lo
        h[nn] = s_lo[nn] / o_hi
    return l, h



from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np
from functools import partial
import itertools
from ..pba.pbox_abc import convert_pbox
from ..pba.aggregation import stacking
from .b2b import b2b
from ..pba.dependency import Dependency
from ..pba.params import Params

if TYPE_CHECKING:
    from ..pba.intervals import Interval
    from ..pba.distributions import Distribution, JointDistribution
    from ..pba.pbox_abc import Pbox


def interval_monte_carlo(
    vars: list[Interval | Distribution | Pbox],
    func: callable,
    interval_strategy,
    n_sam: int,
    dependency: Dependency = None,
    random_state=None,
    side_effects=False,
    **kwargs,
) -> Pbox:
    """Interval Monte Carlo for propagation of pbox

    args:
        vars (list): a list of constructs

        func (callable) : response function. By default, iterable signature is expected.

        interval_strategy (str) :
            strategy for interval discretisation, options include {'direct', 'endpoints', 'subinterval'}

        n_sam (int):
            number of samples for each input

        dependency: dependency structure (e.g. vine copula or archimedean copula

        random_state: random seed for reproducibility

        side_effects (bool): whether return auxiliary outputs (side effects) during propagation
            If true, the alpha-cut samples in the uniform space will be returned as well.
            otherwise, the default is False and only the p-box is returned.


    note:
        When choosing ``interval_strategy``, "direct" requires function signature to take a list of inputs,
        whereas "subinterval" and "endpoints" require the function to have a vectorised signature.

    return:
        Pbox

    example:
        >>> from pyuncertainnumber import pba
        >>> def foo(x): return x[0] ** 3 + x[1] + x[2]
        >>> a = pba.normal([2, 3], [1])
        >>> b = pba.normal([10, 14], [1])
        >>> c = pba.normal([4, 5], [1])
        >>> corre_matrix = np.array([[1, 0.5, 0.3], [0.5, 1, 0.4], [0.3, 0.4, 1]])
        >>> de = pba.Dependency(family='gaussian', corr=corre_matrix)
        >>> mix = interval_monte_carlo(vars=[a,b,c],
        >>> ...       func=foo,
        >>> ...       n_sam=20,
        >>> ...       dependency=de,
        >>> ...       interval_strategy='direct')
    """

    vars = [convert_pbox(v) for v in vars]

    n_sam = int(n_sam) if isinstance(n_sam, float) else n_sam

    b2b_f = partial(b2b, func=func, interval_strategy=interval_strategy, **kwargs)

    if dependency is None:
        ndim = len(vars)
        dependency = Dependency(family="independence", k_dim=ndim)
        """old code for independent case only, which is driven by a cartesian product implementation"""

    prob_proxy_input = dependency.u_sample(n_sam, random_state=random_state)

    # TODO chanege alpha_cut to get outer_approximation
    container = []
    for i, row in enumerate(prob_proxy_input):
        x_domain = [
            v.alpha_cut(a) for v, a in zip(vars, row)
        ]  # yield a list of intervals

        response_y_itvl = b2b_f(x_domain)
        container.append(response_y_itvl)

    if not side_effects:
        return stacking(container)
    else:
        return stacking(container), prob_proxy_input


def slicing(
    vars: list[Distribution | Interval | Pbox],
    func,
    interval_strategy,
    n_slices,
    outer_discretisation=True,
    dependency=None,
    **kwargs,
) -> Pbox:
    """classic slicing algoritm for rigorous propagation of pbox

    args:
        vars (list): list of constructs

        func (callable) : response function

        interval_strategy (str) : strategy for interval discretisation, options include {'direct', 'endpoints', 'subinterval'}

        n_slices: number of slices for each input

        outer_discretisation (bool): whether to use outer discretisation for pbox.
            By default is True for rigorous propagation; however, alpha-cut style interval are also supported.

        dependency: dependency structure (e.g. vine copula or archimedean copula).

    tip:
        Merely independence assumption is supported by now. Other dependency structures are at beta developement now.

    note:
        When choosing ``interval_strategy``, "direct" requires function signature to take a list of inputs,
        whereas "subinterval" and "endpoints" require the function to have a vectorised signature.

    return:
        Pbox

    example:
        >>> from pyuncertainnumber import pba
        >>> def foo(x): return x[0] ** 3 + x[1] + x[2]
        >>> a = pba.normal([2, 3], [1])
        >>> b = pba.normal([10, 14], [1])
        >>> c = pba.normal([4, 5], [1])
        >>> mix = slicing(vars=[a,b,c],
        >>> ...       func=foo,
        >>> ...       n_slices=20,
        >>> ...       interval_strategy='direct')

    """
    p_vars = [convert_pbox(v) for v in vars]

    # if outer_discretisation:
    #     itvs = [p.outer_discretisation(n_slices) for p in p_vars]
    # else:
    #     itvs = [v.discretise(n_slices) for v in p_vars]

    # if len(itvs) == 1:
    #     response_intvl = func(itvs[0])
    #     response_pbox = stacking(response_intvl)
    #     return response_pbox

    b2b_f = partial(b2b, func=func, interval_strategy=interval_strategy, **kwargs)
    # cartesian product of intervals -- original implementation but really slow and not efficient;
    # container = [b2b_f(_item) for _item in itertools.product(*itvs)]

    def make_u_sample(n, num_points):
        # 200 equally spaced points between 0 and 1
        # grid_1d = np.linspace(0.01, 0.99, num_points)

        grid_1d = np.linspace(Params.p_lboundary, Params.p_hboundary, num_points)
        # grid_1d = np.arange(Params.p_lboundary, Params.p_hboundary, 1 / num_points)

        # Create n-dimensional meshgrid
        mesh = np.meshgrid(*([grid_1d] * n), indexing="ij")

        # Stack and reshape to (num_points**n, n)
        u_sample = np.stack(mesh, axis=-1).reshape(-1, n)

        return u_sample

    prob_proxy_input = make_u_sample(n=len(vars), num_points=n_slices)

    container = []
    for i, row in enumerate(prob_proxy_input):
        x_domain = [
            v.alpha_cut(a) for v, a in zip(p_vars, row)
        ]  # yield a list of intervals

        response_y_itvl = b2b_f(x_domain)
        container.append(response_y_itvl)

    # masses = dependency.pdf(prob_proxy_input)

    # return stacking(container, weights=masses)

    return stacking(container)


def double_monte_carlo(
    joint_distribution: Distribution | JointDistribution,
    epistemic_vars: list[Interval],
    n_a: int,
    n_e: int,
    func: callable,
    side_effects=False,
    parallel=False,
) -> tuple[Pbox, list, np.ndarray]:
    """Double-loop Monte Carlo or nested Monte Carlo for mixed uncertainty propagation

    args:
        joint_distribution (Distribution or JointDistribution): an aleatoric sampler based on joint distribution of aleatory variables (or marginal one in 1d case).
            A sampler is basically anything (univariate or multivariate) that has the `sample` interface whereby it can sample a given number of samples.
        epistemic_vars (list): a list epistemic variables in the form of Interval
        n_a (int): number of aleatory samples
        n_e (int): number of epistemic samples
        parallel (Boolean): parallel processing. Only use it for heavy computation (black-box) due to overhead

    hint:
        consider a function mapping f(X) -> y

        - :math:`X` in :math:`R^5` with `n_a=1000`will suggest f(1000, 5)

        - resulting sample array: with `n_e=2`, the response :math:`y` : (n_ep+2, n_a) e.g. (4, 1000)


    side_effects (bool): whether return auxiliary outputs (side effects) during propagation
            If true, the alpha-cut samples in the uniform space will be returned as well.
            otherwise, the default is False and only the p-box is returned.

    return:
        If `side_effects` is True, a tuple containing the following items:
            - a p-box enveloping all the CDFs from the epistemic samples
            - a list of ECDFs for each epistemic sample
            - numpy array of shape ``(n_e+2, n_a)`` as a collection of CDFs for the response
            - the epistemic samples used

        Otherwise, just the p-box.


    note:
        The result array can be interpreted as a collection of CDFs for the response function evaluated at the aleatory samples for each epistemic sample.
        One can further envelope these CDFs into a ``Pbox`` or ``UncertainNumber`` object.

    example:
        >>> from pyuncertainnumber import pba
        >>> # vectorised function signature with matrix input (2D np.ndarray)
        >>> def foo_vec(x):
        ...     return x[:, 0] ** 3 + x[:, 1] + x[:, 2] + x[:, 3]

        >>> dist_a = pba.Distribution('gaussian', (5, 1))
        >>> dist_b = pba.Distribution('uniform', (2, 3))
        >>> c = pba.Dependency('gaussian', params=0.8)
        >>> joint_dist = pba.JointDistribution(copula=c, marginals=[dist_a, dist_b])

        >>> xe1 = pba.I(1, 2)
        >>> xe2 = pba.I(3, 4)

        >>> t = double_monte_carlo(
        ...     joint_distribution=joint_dist,
        ...     epistemic_vars=[xe1, xe2],
        ...     n_a=20,
        ...     n_e=3,
        ...     func=foo_vec
        ... )
    """
    # from epistemic vars into vec interval object
    from pyuncertainnumber import make_vec_interval, parse_bounds
    from pyuncertainnumber.pba.distributions import ECDF
    from pyuncertainnumber import envelope

    v = parse_bounds(epistemic_vars)
    # lhs sample array on epistemic variables
    epistemic_points = v.endpoints_lhs_sample(n_e)

    def evaluate_func_on_e(e, n_a, func):
        """propagate wrt one point in the epistemic space

        args:
            e: one point in the epistemic space
            n_a: number of aleatory samples
            func: function to be evaluated

        note:
            by default, aleatory variable are put in front of the epistemic ones
        """
        xa_samples = joint_distribution.sample(n_a).reshape(-1, 1)

        E = np.tile(e, (n_a, 1))
        X_input = np.concatenate((xa_samples, E), axis=1)
        return func(X_input)

    p_func = partial(evaluate_func_on_e, n_a=n_a, func=func)
    container = map(p_func, epistemic_points)
    response = np.squeeze(np.stack(list(container), axis=0))  # (n_e, n_a)

    many_ecdfs = [ECDF(r) for r in response]
    env_pbox = envelope(*many_ecdfs, output_type="pbox")

    if not side_effects:
        return env_pbox
    else:
        return env_pbox, many_ecdfs, response, epistemic_points


def bi_imc(x, y, func, dependency=None, n_sam=100):
    """Bivariate interval monte carlo for convenience

    args:
        x, y (Pbox) : Pbox
        func: callable which takes vector-type of inputs
        dependency: dependency structure (regular copula)
    """
    from scipy.stats import qmc

    # from pyuncertainnumber.pba.aggregation import stacking

    alpha = np.squeeze(qmc.LatinHypercube(d=1).random(n=n_sam))
    x_i = x.alpha_cut(alpha)
    y_i = y.alpha_cut(alpha)

    container = [func(_item) for _item in itertools.product(x_i, y_i)]
    return stacking(container)



from scipy.stats import ecdf
import os
import pathlib
import re
import ast
import json
import dataclasses
import numpy as np
from pymongo import MongoClient
import matplotlib.pyplot as plt
import scipy.stats as sps


# TODO create a defending mechanism for parsing '[15+-10%]' as only '[15 +- 10%]' works now


def save2Vis(figname):
    """a shortcut function to save plot to visualization dir

    Note
    ----

    We simply assume that every repo will have a 'visulizations'
    dir under the root directory
    """

    axe = plt.gca()
    plt.savefig(f"{figname}.png", format="png", dpi=300, bbox_inches="tight")


def tranform_ecdf(s, display=False, **kwargs):
    """plot the CDF return the quantile

    args:
        s: sample
    """
    sth = ecdf(s)
    if display:
        fig, ax = plt.subplots()
        # ax.plot(x_support, p_values, color='g')
        ax.step(
            sth.cdf.quantiles,
            sth.cdf.probabilities,
            color="red",
            zorder=10,
            where="post",
            **kwargs,
        )
        return sth.cdf.quantiles, ax
    else:
        return sth.cdf.quantiles


def pl_pcdf(
    dist: type[sps.rv_continuous | sps.rv_discrete], ax=None, title=None, **kwargs
):
    from ..pba.params import Params

    """plot CDF from parametric distribution objects"""

    if ax is None:
        fig, ax = plt.subplots()
    if title is not None:
        ax.set_title(title)

    x_values = dist.ppf(Params.p_values)

    ax.plot(x_values, Params.p_values, **kwargs)
    ax.set_xlabel(r"$x$")
    ax.set_ylabel(r"$\Pr(X \leq x)$")
    return ax


def pl_ecdf(s, ax=None, return_value=False, **kwargs):
    """plot the empirical CDF given samples

    args:
        s (array-like): sample which can be either raw data
            or deviates as a representation of dist construct
    """
    sth = ecdf(s)
    if ax is None:
        fig, ax = plt.subplots()
    # ax.plot(x_support, p_values, color='g')
    ax.step(sth.cdf.quantiles, sth.cdf.probabilities, where="post", **kwargs)
    ax.set_xlabel(r"$x$")
    ax.set_ylabel(r"$\Pr(X \leq x)$")
    if not return_value:
        return ax
    else:
        return ax, sth.cdf.quantiles, sth.cdf.probabilities


def to_database(dict_list, db_name, col_name):

    json_rep = json.dumps(dict_list, cls=UNEncoder)
    dict_rep = json.loads(json_rep)

    with MongoClient() as client:
        db = client[db_name]

        # collection
        a_collection = db[col_name]

        # insert documents
        new_result = a_collection.insert_many(dict_rep)
        return new_result


def cd_root_dir(depth=0):
    # change directory to the path of the root directory of the project

    ref_path = os.path.abspath("")
    ref_path = pathlib.Path(ref_path).resolve().parents[depth]
    os.chdir(ref_path)
    print("current directory:", os.getcwd())


def sgnumber(
    user_input: str,
) -> list:
    """significant digits representation

    number plus or minus its significant-digit imprecision
    """
    user_input = user_input.strip().lower()
    tens = "0"
    if "e" in user_input:
        mantissa, tens = user_input.split("e", 1)
    else:
        mantissa = user_input
    if "." in mantissa:
        j = len(mantissa.split(".")[1])
    # else: j = len(mantissa.split('0', 1)) - len(mantissa) + 1
    else:
        j = len(mantissa.rstrip("0")) - len(mantissa)
    pm = 10 ** (-j) * 10 ** int(tens) / 2
    # print('input:',user_input,', mantissa:',mantissa, ', j:',j, ', tens:',tens, ', pm:',pm)
    return [float(user_input) - pm, float(user_input) + pm]


def initial_list_checking(text):
    """detects if a string representation of a list"""

    try:
        return ast.literal_eval(text)
    except:
        # print(error)
        # print("Not a list-like string representation")
        pass


def bad_list_checking(text):
    """detects if a syntactically wrong specification of a list"""

    flag = text.startswith("[") & text.endswith("]")
    # if flag:
    #     print("Wrong spec of a list repre")
    # else:
    #     print("Not even a list")
    return flag


def PlusMinus_parser(txt):

    flag = "+-" in txt
    if flag:
        # print("Contains '+-' ergo initiate using mid range style")
        return True
        # txt_list = list(txt)
    # return txt_list


# def deciper_num_from_string():


def parser4(text):

    # do an extra step of scraping the '[' and ']'
    if bad_list_checking(text):
        subtexts = text.strip("[]").split()
    else:
        subtexts = text.split()
    return [int(s) for s in subtexts if s.isdigit()]


def percentage_finder(txt):
    pctg = re.findall("\d*%", txt)
    # return pctg
    if pctg:
        return True
    else:
        return False


def percentage_converter(txt):
    """convert a percentage into a float number

    note:
        force only 1 percentage
    """
    # return re.findall(r'(\d+(\.\d+)?%)', txt)

    pctg = re.findall("\d*%", txt)
    return float(pctg[0].strip("%")) / 100


class EnhancedJSONEncoder(json.JSONEncoder):
    """a template for jsonify general (dataclass) object

    #TODO Interval object in not json serializable
    """

    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        return super().default(o)


class PBAEncoder(json.JSONEncoder):
    """a bespoke JSON encoder for the PBA object"""

    def default(self, o):
        # if any(isinstance(value, np.ndarray) for value in o.__dict__.values()):
        # TODO to use __slot__ later on to save disk space
        removed_dict = o.__dict__.copy()
        entries_to_remove(remove_entries=["left", "right"], the_dict=removed_dict)
        return removed_dict


class UNEncoder(json.JSONEncoder):
    """a bespoke JSON encoder for the UncertainNumber object

    note:
        - Currently I'm treating the JSON data represent of a UN object
        the same as the __repr__ method. But this can be changed later on to
        show more explicitly the strucutre of pbox or distribution
        # TODO prettify the JSON output to be explicit
        e.g. 'essence': 'interval', 'interval_initialisation': [2, 3] to shown as 'interval' with lower end and upper end
        distribution to shown as the type and parameters; e.g. 'distribution': 'normal', 'parameters': [2, 3]
    """

    def default(self, o):
        # if any(isinstance(value, np.ndarray) for value in o.__dict__.values()):
        # TODO to use __slot__ later on to save disk space
        copy_dict = o.__dict__.copy()

        return get_concise_repr(copy_dict)


def get_concise_repr(a_dict):
    # remove None fields
    Noneremoved_dict = {k: v for k, v in a_dict.items() if v is not None}

    # remove some unwanted fields (keys)
    entries_to_remove(
        remove_entries=[
            "distribution_parameters",
            "pbox_parameters",
            "bounds",
            "masses",
            "_units",
            "_physical_quantity",
            "_skip_construct_init",
            "_UnitsRep",
            "_math_object",
            "deter_value_rep",
            "_Q",
            "p_flag",
        ],
        the_dict=Noneremoved_dict,
    )
    return Noneremoved_dict


def array2list(a_dict):
    """convert an array from a dictionary into a list"""

    return {
        k: arr.tolist() if isinstance(arr, np.ndarray) else arr
        for k, arr in a_dict.items()
    }


def entries_to_remove(remove_entries, the_dict):

    for key in remove_entries:
        if key in the_dict:
            del the_dict[key]



from __future__ import annotations
import numpy as np
import matplotlib.pyplot as plt
from numpy.typing import ArrayLike

"""
--------------------------
Editted by Leslie Feb 2024 
MIT License
--------------------------
"""


def plot_intervals(x, y_i, **kwargs):
    """plot intervals vertically

    args:
        x: array-like precise values
            x-axis coordinates
        y_i: array-like Interval objects
            array of intervals
    """

    fig, ax = plt.subplots()

    def basic_plot(x, y_i, **kwargs):
        ax.plot([x, x], [y_i.hi, y_i.lo], "blue", **kwargs)
        if np.any(y_i.lo == y_i.hi):
            sc_x = x[y_i.lo == y_i.hi]
            sc_y = y_i[y_i.lo == y_i.hi].lo
            ax.scatter(sc_x, sc_y, c="blue", **kwargs)

    if len(x.shape) > 1:
        for xx, interval in zip(x, y_i):
            basic_plot([xx, xx], [interval.hi, interval.lo])
    else:
        basic_plot(x, y_i)
    return ax


def plot_lower_bound(x: ArrayLike, y_i: ArrayLike, **kwargs):
    """plot lower bound of intervals

    args:
        x (ArrayLike): x-axis coordinates
        y_i (ArrayLike): array of intervals
    """

    fig, ax = plt.subplots()
    ax.scatter(x, y_i.lo, label="lower bound", **kwargs)
    ax.legend()


# TODO: finsih the logic for when probability is given
def plot_mul_intervals(vec_interval, probability=None, ax=None):
    """plot multiple intervals horizontally in the 0-1 probability range

    args:
        vec_interval (Interval): a vector interval object
    """
    if (
        probability is None
    ):  # this will plot the intervals across the 0-1 probability axis
        from ..utils import equi_selection

        p_values = np.linspace(0, 1, 200)
        pp = equi_selection(p_values, len(vec_interval))

        if ax is None:
            fig, ax = plt.subplots()

        for i, intl in enumerate(vec_interval):  # horizontally plot the interval
            ax.plot([intl.lo, intl.hi], [pp[i], pp[i]], label=f"Interval {i+1}")
        ax.margins(x=0.1, y=0.1)
        return ax
    else:  # when probability is given
        print('unfinished function "plot_mul_intervals" with given probability')



# read version from installed package
from importlib.metadata import version

__version__ = version(__name__)


from pyuncertainnumber.characterisation.uncertainNumber import UncertainNumber as UN
from pyuncertainnumber.characterisation.uncertainNumber import *

# * --------------------- pba ---------------------*#
import pyuncertainnumber.pba as pba
from pyuncertainnumber.pba.pbox_free import *
from pyuncertainnumber.characterisation.stats import fit


# * --------------------- Pbox ---------------------*#
from pyuncertainnumber.pba.pbox_abc import Pbox, Staircase

# * --------------------- Interval ---------------------*#
from pyuncertainnumber.pba.intervals.number import Interval
from pyuncertainnumber.pba.intervals.intervalOperators import (
    make_vec_interval,
    parse_bounds,
)
from pyuncertainnumber.pba.intervals import intervalise
from pyuncertainnumber.propagation.helper import EpistemicDomain


# * --------------------- hedge---------------------*#
from pyuncertainnumber.nlp.language_parsing import hedge_interpret


# * --------------------- cbox ---------------------*#
from pyuncertainnumber.pba.cbox import infer_cbox, infer_predictive_distribution


# * --------------------- DempsterShafer ---------------------*#
from pyuncertainnumber.pba.dss import dempstershafer_element, DempsterShafer


# * --------------------- Dependency ---------------------*#
from pyuncertainnumber.pba.dependency import Dependency

# * --------------------- Characterisation ---------------------*#
from pyuncertainnumber.pba.pbox_free import KS_bounds


# * ---------------------  aggregation ---------------------*#
from pyuncertainnumber.pba.aggregation import *

# * ---------------------  propagation ---------------------*#
from pyuncertainnumber.propagation.b2b import b2b
from pyuncertainnumber.propagation.p import Propagation
from pyuncertainnumber.propagation.taylor_expansion import taylor_expansion_method
from pyuncertainnumber.propagation.mixed_up import (
    interval_monte_carlo,
    slicing,
    double_monte_carlo,
)


# * --------------------- validation ---------------------*#
from .pba.core import area_metric

# * ---------------------  utils ---------------------*#
from pyuncertainnumber.gutils import inspect_un



from __future__ import annotations
import functools
from .pbox_abc import Pbox, Leaf
import scipy.stats as sps
import numpy as np
import itertools
from .params import Params
from typing import *
from warnings import *
from .intervals.intervalOperators import wc_scalar_interval

if TYPE_CHECKING:
    from .pbox_abc import Pbox
    from ..pba.intervals.number import Interval


""" parametric pboxes"""


def makePbox(func) -> Pbox:
    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):
        family_str = func(*args, **kwargs)
        return _bound_pcdf(family_str, *args, **kwargs)

    return wrapper_decorator


# top-level
def _bound_pcdf(dist_family, *args, **kwargs):
    """bound the parametric CDF

    note:
        - top-level implemenatation
        - only support fully bounded parameters
    """
    from .pbox_abc import Leaf

    Left, Right, mean, var = _parametric_bounds_array(dist_family, *args, **kwargs)
    return Leaf(
        left=Left, right=Right, shape=dist_family, dist_params=args, mean=mean, var=var
    )


def _parametric_bounds_array(dist_family, *args, **kwargs):
    """from parametric distribution specification to define the lower and upper bound of the p-box

    args:
        dist_family: (str) the name of the distribution
        *args : several parameter (interval or list)
        **kwargs : scale parameters (interval or list)

    note:
        - middle level implementation

    """

    from .distributions import named_dists
    from .intervals.number import Interval as I

    i_args = [wc_scalar_interval(b) for b in args]

    if kwargs:
        kw_args = [wc_scalar_interval(v) for v in kwargs.values()]
        i_args = I(0.0, 0.0)
        new_args = itertools.product(i_args.val, *[i.to_numpy() for i in kw_args])
    else:
        new_args = itertools.product(*[i.to_numpy() for i in i_args])

    dist = named_dists[dist_family]
    g1, g2 = itertools.tee(new_args, 2)
    bounds = [dist.ppf(Params.p_values, *a) for a in g1]
    stats = [dist.stats(*a, moments="mv") for a in g2]

    means, vars_ = zip(*stats)
    mean = I(min(means), max(means))
    var = I(min(vars_), max(vars_))

    Left = np.min(bounds, axis=0)
    Right = np.max(bounds, axis=0)

    return Left, Right, mean, var


# * ---------------------supported distributional pboxes ---------------------#


@makePbox
def norm(*args):
    return "norm"


@makePbox
def lognormal(*args):
    return "lognormal"


@makePbox
def alpha(*args):
    return "alpha"


@makePbox
def anglit(*args):
    return "anglit"


@makePbox
def argus(*args):
    return "argus"


@makePbox
def arcsine(*args):
    return "arcsine"


@makePbox
def beta(*args):
    return "beta"


@makePbox
def betaprime(*args):
    return "betaprime"


@makePbox
def bradford(*args):
    return "bradford"


@makePbox
def burr(*args):
    return "burr"


@makePbox
def burr12(*args):
    return "burr12"


@makePbox
def cauchy(*args):
    return "cauchy"


@makePbox
def chi(*args):
    return "chi"


@makePbox
def chi2(*args):
    return "chi2"


@makePbox
def cosine(*args):
    return "cosine"


@makePbox
def crystalball(*args):
    return "crystalball"


@makePbox
def dgamma(*args):
    return "dgamma"


@makePbox
def dweibull(*args):
    return "dweibull"


@makePbox
def erlang(*args):
    return "erlang"


@makePbox
def exponnorm(*args):
    return "exponnorm"


@makePbox
def exponential(*args, **kwargs):
    """The default p-box constructor for the exponential distribution with scale parameterisation

    note:
        scale parameterisation due to scipy.stats. Note that the "scale" argument is a must.
        There is an "exponential_by_lambda" constructor which uses the rate parameterisation.

    example:
        >>> pba.pbox_parametric.exponential(scale=[1, 2])
    """
    return "exponential"


@makePbox
def exponweib(*args):
    return "exponweib"


@makePbox
def exponpow(*args):
    return "exponpow"


@makePbox
def f(*args):
    return "f"


@makePbox
def fatiguelife(*args):
    return "fatiguelife"


@makePbox
def fisk(*args):
    return "fisk"


@makePbox
def foldcauchy(*args):
    return "foldcauchy"


def foldnorm(mu, s, steps=Params.steps):

    x = np.linspace(0.0001, 0.9999, steps)
    if mu.__class__.__name__ != "wc_scalar_interval":
        mu = wc_scalar_interval(mu)
    if s.__class__.__name__ != "wc_scalar_interval":
        s = wc_scalar_interval(s)

    new_args = [
        [mu.lo() / s.lo(), 0, s.lo()],
        [mu.hi() / s.lo(), 0, s.lo()],
        [mu.lo() / s.hi(), 0, s.hi()],
        [mu.hi() / s.hi(), 0, s.hi()],
    ]

    bounds = []

    mean_hi = -np.inf
    mean_lo = np.inf
    var_lo = np.inf
    var_hi = 0

    for a in new_args:

        bounds.append(sps.foldnorm.ppf(x, *a))
        bmean, bvar = sps.foldnorm.stats(*a, moments="mv")

        if bmean < mean_lo:
            mean_lo = bmean
        if bmean > mean_hi:
            mean_hi = bmean
        if bvar > var_hi:
            var_hi = bvar
        if bvar < var_lo:
            var_lo = bvar

    Left = [min([b[i] for b in bounds]) for i in range(steps)]
    Right = [max([b[i] for b in bounds]) for i in range(steps)]

    var = wc_scalar_interval(np.float64(var_lo), np.float64(var_hi))
    mean = wc_scalar_interval(np.float64(mean_lo), np.float64(mean_hi))

    Left = np.array(Left)
    Right = np.array(Right)

    return Pbox(
        Left,
        Right,
        steps=steps,
        shape="foldnorm",
        mean_left=mean.left,
        mean_right=mean.right,
        var_left=var.left,
        var_right=var.right,
    )


@makePbox
def genlogistic(*args):
    return "genlogistic"


@makePbox
def gennorm(*args):
    return "gennorm"


@makePbox
def genpareto(*args):
    return "genpareto"


@makePbox
def genexpon(*args):
    return "genexpon"


@makePbox
def genextreme(*args):
    return "genextreme"


@makePbox
def gausshyper(*args):
    return "gausshyper"


@makePbox
def gamma(*args):
    return "gamma"


@makePbox
def gengamma(*args):
    return "gengamma"


@makePbox
def genhalflogistic(*args):
    return "genhalflogistic"


@makePbox
def geninvgauss(*args):
    return "geninvgauss"


@makePbox
def gompertz(*args):
    return "gompertz"


@makePbox
def gumbel_r(*args):
    return "gumbel_r"


@makePbox
def gumbel_l(*args):
    return "gumbel_l"


@makePbox
def halfcauchy(*args):
    return "halfcauchy"


@makePbox
def halflogistic(*args):
    return "halflogistic"


@makePbox
def halfnorm(*args):
    return "halfnorm"


@makePbox
def halfgennorm(*args):
    return "halfgennorm"


@makePbox
def hypsecant(*args):
    return "hypsecant"


@makePbox
def invgamma(*args):
    return "invgamma"


@makePbox
def invgauss(*args):
    return "invgauss"


@makePbox
def invweibull(*args):
    return "invweibull"


@makePbox
def irwinhall(*args):
    return "irwinhall"


@makePbox
def jf_skew_t(*args):
    return "jf_skew_t"


@makePbox
def johnsonsb(*args):
    return "johnsonsb"


@makePbox
def johnsonsu(*args):
    return "johnsonsu"


@makePbox
def kappa4(*args):
    return "kappa4"


@makePbox
def kappa3(*args):
    return "kappa3"


@makePbox
def ksone(*args):
    return "ksone"


@makePbox
def kstwo(*args):
    return "kstwo"


@makePbox
def kstwobign(*args):
    return "kstwobign"


@makePbox
def laplace(*args):

    return "laplace"


@makePbox
def laplace_asymmetric(*args):
    return "laplace_asymmetric"


@makePbox
def levy(*args):
    return "levy"


@makePbox
def levy_l(*args):
    return "levy_l"


@makePbox
def levy_stable(*args):
    return "levy_stable"


@makePbox
def logistic(*args):
    return "logistic"


@makePbox
def loggamma(*args):
    return "loggamma"


@makePbox
def loglaplace(*args):
    return "loglaplace"


@makePbox
def loguniform(*args):
    return "loguniform"


@makePbox
def lomax(*args):
    return "lomax"


@makePbox
def maxwell(*args):
    return "maxwell"


@makePbox
def mielke(*args):
    return "mielke"


@makePbox
def moyal(*args):
    return "moyal"


@makePbox
def nakagami(*args):
    return "nakagami"


@makePbox
def ncx2(*args):
    return "ncx2"


@makePbox
def ncf(*args):
    return "ncf"


@makePbox
def nct(*args):
    return "nct"


@makePbox
def norminvgauss(*args):
    return "norminvgauss"


@makePbox
def pareto(*args):
    return "pareto"


@makePbox
def pearson3(*args):
    return "pearson3"


@makePbox
def powerlaw(*args):
    return "powerlaw"


@makePbox
def powerlognorm(*args):
    return "powerlognorm"


@makePbox
def powernorm(*args):
    return "powernorm"


@makePbox
def rdist(*args):
    return "rdist"


@makePbox
def rayleigh(*args, **kwargs):
    return "rayleigh"


@makePbox
def rel_breitwigner(*args):
    return "rel_breitwigner"


@makePbox
def rice(*args):
    return "rice"


@makePbox
def recipinvgauss(*args):
    return "recipinvgauss"


@makePbox
def semicircular(*args):
    return "semicircular"


@makePbox
def skewcauchy(*args):
    return "skewcauchy"


@makePbox
def skewnorm(*args):
    return "skewnorm"


@makePbox
def studentized_range(*args):
    return "studentized_range"


@makePbox
def t(*args):
    return "t"


@makePbox
def trapezoid(*args):
    return "trapezoid"


@makePbox
def triang(*args):
    return "triang"


@makePbox
def truncweibull_min(*args):
    return "truncweibull_min"


@makePbox
def tukeylambda(*args):
    return "tukeylambda"


def uniform_sps(*args):
    return "uniform"


@makePbox
def vonmises(*args):
    return "vonmises"


@makePbox
def vonmises_line(*args):
    return "vonmises_line"


@makePbox
def wald(*args):
    return "wald"


@makePbox
def weibull_min(*args):
    return "weibull_min"


@makePbox
def weibull_max(*args):
    return "weibull_max"


@makePbox
def wrapcauchy(*args):
    return "wrapcauchy"


# *---------------------some special ones ---------------------*#


def lognormal_weird(
    mean,
    var,
    steps=Params.steps,
):
    """p-box for the lognormal distribution

    *Note: the parameters used are the mean and variance of the lognormal distribution

    not the mean and variance of the underlying normal*
    See:
    `[1]<https://en.wikipedia.org/wiki/Log-normal_distribution#Generation_and_parameters>`
    `[2]<https://stackoverflow.com/questions/51906063/distribution-mean-and-standard-deviation-using-scipy-stats>`


    Parameters
    ----------
    mean :
        mean of the lognormal distribution
    var :
        variance of the lognormal distribution

    Returns
    -------
    Pbox

    """

    x = np.linspace(0.001, 0.999, Params.steps)
    mean, var = wc_scalar_interval(mean), wc_scalar_interval(var)

    def __lognorm(mean, var):
        sigma = np.sqrt(np.log1p(var / mean**2))
        mu = np.log(mean) - 0.5 * sigma * sigma
        return sps.lognorm(sigma, loc=0, scale=np.exp(mu))

    bound0 = __lognorm(mean.left, var.left).ppf(x)
    bound1 = __lognorm(mean.right, var.left).ppf(x)
    bound2 = __lognorm(mean.left, var.right).ppf(x)
    bound3 = __lognorm(mean.right, var.right).ppf(x)

    Left = [min(bound0[i], bound1[i], bound2[i], bound3[i]) for i in range(steps)]
    Right = [max(bound0[i], bound1[i], bound2[i], bound3[i]) for i in range(steps)]

    Left = np.array(Left)
    Right = np.array(Right)
    return Leaf(left=Left, right=Right, steps=steps, shape="lognormal")


def uniform(a, b, steps=Params.steps):
    """special case of Uniform distribution as
    Scipy has an unbelivably strange parameterisation than common sense

    args:
        - a: (float) lower endpoint
        - b: (float) upper endpoints
    """

    # loc, scale = uniform_reparameterisation(a,  b)
    # return uniform_sps(loc, scale)

    a, b = [wc_scalar_interval(arg) for arg in [a, b]]

    Left = np.linspace(a.left, b.left, steps)
    Right = np.linspace(a.right, b.right, steps)

    return Leaf(
        left=Left,
        right=Right,
        steps=steps,
        shape="uniform",
    )


def exponential_by_lambda(lamb: list | Interval) -> Pbox:
    """Bespoke p-box constructor for the exponential distribution

    args:
        - lamb: (list or Interval) the rate parameter of the exponential distribution
    """
    from .distributions import expon_sane
    from .pbox_abc import Staircase

    interval_lambda = wc_scalar_interval(lamb)

    a = expon_sane(interval_lambda.lo)
    b = expon_sane(interval_lambda.hi)

    a_quantile = a.ppf(Params.p_values)
    b_quantile = b.ppf(Params.p_values)

    try:
        p = Staircase(left=a_quantile, right=b_quantile)
    except Exception as e:
        p = Staircase(left=b_quantile, right=a_quantile)
    return p


def trapz(a, b, c, d, steps=Params.steps):
    if a.__class__.__name__ != "wc_scalar_interval":
        a = wc_scalar_interval(a)
    if b.__class__.__name__ != "wc_scalar_interval":
        b = wc_scalar_interval(b)
    if c.__class__.__name__ != "wc_scalar_interval":
        c = wc_scalar_interval(c)
    if d.__class__.__name__ != "wc_scalar_interval":
        d = wc_scalar_interval(d)

    x = np.linspace(0.0001, 0.9999, steps)
    left = sps.trapz.ppf(
        x, *sorted([b.lo() / d.lo(), c.lo() / d.lo(), a.lo(), d.lo() - a.lo()])
    )
    right = sps.trapz.ppf(
        x, *sorted([b.hi() / d.hi(), c.hi() / d.hi(), a.hi(), d.hi() - a.hi()])
    )

    return Pbox(left, right, steps=steps, shape="trapz")


def weibull(*args, steps=Params.steps):

    wm = weibull_max(*args)
    wl = weibull_min(*args)

    return Pbox(left=wl.left, right=wm.right)


# Other distributions
def KM(k, m, steps=Params.steps):
    with catch_warnings():
        simplefilter("ignore")
        return beta(
            wc_scalar_interval(k, k + 1), wc_scalar_interval(m, m + 1), steps=steps
        )


def KN(k, n, steps=Params.steps):
    return KM(k, n - k, steps=steps)


# *---------------------discrete distributions---------------------*#


@makePbox
def bernoulli(*args):
    return "bernoulli"


@makePbox
def betabinom(*args):
    return "betabinom"


@makePbox
def betanbinom(*args):
    return "betanbinom"


@makePbox
def binom(*args):
    return "binom"


@makePbox
def boltzmann(*args):
    return "boltzmann"


@makePbox
def dlaplace(*args):
    return "dlaplace"


@makePbox
def geom(*args):
    return "geom"


@makePbox
def hypergeom(*args):
    return "hypergeom"


@makePbox
def logser(*args):
    return "logser"


@makePbox
def nbinom(*args):
    return "nbinom"


@makePbox
def nchypergeom_fisher(*args):
    return "nchypergeom_fisher"


@makePbox
def nchypergeom_wallenius(*args):
    return "nchypergeom_wallenius"


@makePbox
def nhypergeom(*args):
    return "nhypergeom"


@makePbox
def planck(*args):
    return "planck"


@makePbox
def poisson(*args):
    return "poisson"


@makePbox
def randint(*args):
    return "randint"


@makePbox
def skellam(*args):
    return "skellam"


@makePbox
def yulesimon(*args):
    return "yulesimon"


@makePbox
def zipf(*args):
    return "zipf"


@makePbox
def zipfian(*args):
    return "zipfian"


# *---------------------aliases---------------------*#
normal = norm
gaussian = norm
# lognormal = lognormal_weird

# *---------------------named pboxes for UN ---------------------*#
named_pbox = {
    "alpha": alpha,
    "anglit": anglit,
    "arcsine": arcsine,
    "argus": argus,
    "beta": beta,
    "betaprime": betaprime,
    "bradford": bradford,
    "burr": burr,
    "burr12": burr12,
    "cauchy": cauchy,
    "chi": chi,
    "chi2": chi2,
    "cosine": cosine,
    "crystalball": crystalball,
    "dgamma": dgamma,
    "dweibull": dweibull,
    "erlang": erlang,
    "exponential": exponential,
    "exponnorm": exponnorm,
    "exponweib": exponweib,
    "exponpow": exponpow,
    "f": f,
    "fatiguelife": fatiguelife,
    "fisk": fisk,
    "foldcauchy": foldcauchy,
    "foldnorm": foldnorm,
    # 'frechet_r' : frechet_r,
    # 'frechet_l' : frechet_l,
    "genlogistic": genlogistic,
    "gennorm": gennorm,
    "genpareto": genpareto,
    "genexpon": genexpon,
    "genextreme": genextreme,
    "gausshyper": gausshyper,
    "gamma": gamma,
    "gengamma": gengamma,
    "genhalflogistic": genhalflogistic,
    "geninvgauss": geninvgauss,
    # 'gibrat' : gibrat,
    "gompertz": gompertz,
    "gumbel_r": gumbel_r,
    "gumbel_l": gumbel_l,
    "halfcauchy": halfcauchy,
    "halflogistic": halflogistic,
    "halfnorm": halfnorm,
    "halfgennorm": halfgennorm,
    "hypsecant": hypsecant,
    "invgamma": invgamma,
    "invgauss": invgauss,
    "invweibull": invweibull,
    "johnsonsb": johnsonsb,
    "johnsonsu": johnsonsu,
    "kappa4": kappa4,
    "kappa3": kappa3,
    "ksone": ksone,
    "kstwobign": kstwobign,
    "laplace": laplace,
    "levy": levy,
    "levy_l": levy_l,
    "levy_stable": levy_stable,
    "logistic": logistic,
    "loggamma": loggamma,
    "loglaplace": loglaplace,
    "lognormal": lognormal,
    "loguniform": loguniform,
    "lomax": lomax,
    "maxwell": maxwell,
    "mielke": mielke,
    "moyal": moyal,
    "nakagami": nakagami,
    "ncx2": ncx2,
    "ncf": ncf,
    "nct": nct,
    "norm": norm,
    "normal": norm,
    "gaussian": norm,
    "norminvgauss": norminvgauss,
    "pareto": pareto,
    "pearson3": pearson3,
    "powerlaw": powerlaw,
    "powerlognorm": powerlognorm,
    "powernorm": powernorm,
    "rdist": rdist,
    "rayleigh": rayleigh,
    "rice": rice,
    "recipinvgauss": recipinvgauss,
    "semicircular": semicircular,
    "skewnorm": skewnorm,
    "t": t,
    "trapz": trapz,
    "triang": triang,
    "tukeylambda": tukeylambda,
    "uniform": uniform,
    "vonmises": vonmises,
    "vonmises_line": vonmises_line,
    "wald": wald,
    "weibull_min": weibull_min,
    "weibull_max": weibull_max,
    "wrapcauchy": wrapcauchy,
    "bernoulli": bernoulli,
    "betabinom": betabinom,
    "binom": binom,
    "boltzmann": boltzmann,
    "dlaplace": dlaplace,
    "geom": geom,
    "hypergeom": hypergeom,
    "logser": logser,
    "nbinom": nbinom,
    "planck": planck,
    "poisson": poisson,
    "randint": randint,
    "skellam": skellam,
    "zipf": zipf,
    "yulesimon": yulesimon,
}



from __future__ import annotations
from typing import TYPE_CHECKING
import functools
import numpy as np
from pyuncertainnumber import pba
from scipy.stats import beta, t, gamma, betabinom, nbinom
from .params import Params
from .intervals import Interval
import scipy
from .cbox_constructor import cbox_from_extredists, cbox_from_pseudosamples
from .constructors import pbox_from_pseudosamples
from .pbox_abc import pbox_from_extredists
import pyuncertainnumber.pba.distributions as d

if TYPE_CHECKING:
    from .cbox_constructor import Cbox


def interval_measurements(func):
    """decorator for incorporating interval valued data"""

    # TODO not finished yet
    @functools.wraps(func)
    def imprecise_measurements_wrapper(x, **kwargs):
        if isinstance(x, (list, np.ndarray)):
            conf_dist, params = func(x, **kwargs)
            return cbox_from_extredists(
                conf_dist,
                shape="t",
                extre_bound_params=(params["loc"], params["scale"]),
            )

        elif isinstance(x, Interval):
            cd_lo, params_lo = func(x.lo, **kwargs)
            cd_hi, params_hi = func(x.hi, **kwargs)

            def get_interval_params():
                pass

            return cbox_from_extredists([cd_lo, cd_hi])

    return imprecise_measurements_wrapper


def infer_cbox(family: str, data, **args) -> Cbox:
    """top-level call signature to infer a c-box given data and family, plus rarely additional kwargs

    notes:
        - data (list): a list of data samples, e.g. [2]
        - additina kwargs such as N for binomial family

    example:
        >>> infer_cbox('binomial', data=[2], N=10)
    """
    if isinstance(named_cbox.get(family), dict):
        return {k: v(data, **args) for (k, v) in named_cbox.get(family).items()}
    return named_cbox.get(family)(data, **args)


def infer_predictive_distribution(family: str, data, **args):
    """top-level call for the next value predictive distribution"""
    return named_nextvalue.get(family)(data, **args)


# * ---------------------Bernoulli---------------------*#


def CBbernoulli_p(x):
    n = len(x)
    k = sum(x)
    l_b_params = (k, n - k + 1)
    r_b_params = (k + 1, n - k)
    cdfs = (beta(*l_b_params), beta(*r_b_params))
    return cbox_from_extredists(
        cdfs, shape="beta", extre_bound_params=[l_b_params, r_b_params]
    )


# nextvalue


def CBbernoulli(x):
    n = len(x)
    k = sum(x)
    return pba.bernoulli(np.array([k, k + 1]) / (n + 1))


# * ---------------------binomial---------------------*#
# x[i] ~ binomial(N, p), for known N, x[i] is a nonnegative integer less than or equal to N
def CBbinomial_p(x, N):
    """cbox for Bionomial parameter

    args:
        x (list or int): sample data as in a list of success or number of success or
            a single int as the number of success k
        N (int): number of trials

    note:
        x[i] ~ binomial(N, p), for unknown p, x[i] is a nonnegative integer
        but x is a int number, it suggests the number of success as `k`.

    return:
        cbox: cbox object
    """
    if isinstance(x, int):
        x = [x]
    n = len(x)  # size
    k = sum(x)
    l_b_params = (k, n * N - k + 1)
    r_b_params = (k + 1, n * N - k)
    cdfs = (beta(*l_b_params), beta(*r_b_params))
    return cbox_from_extredists(
        cdfs, shape="beta", extre_bound_params=[l_b_params, r_b_params]
    )


def CBbinomial(x, N):
    if isinstance(x, int):
        x = [x]
    n = len(x)
    k = sum(x)
    cdfs = (betabinom(N, k, n * N - k + 1), betabinom(N, k + 1, n * N - k))
    return pbox_from_extredists(cdfs, shape="betanomial")


# * ---------------------binomialnp---------------------*#
# TODO not done yet
# x[i] ~ binomial(N, p), for unknown N, x[i] is a nonnegative integer
# two parameter version (N, p)
# see https://sites.google.com/site/cboxbinomialnp/
def nextvalue_binomialnp(x):
    pass


def parameter_binomialnp_n(x):
    pass


def parameter_binomialnp_p(x):
    pass


# * ---------------------Poisson---------------------*#
# x[i] ~ Poisson(parameter), x[i] is a nonnegative integer


def CBpoisson_lambda(x):
    n = len(x)
    k = sum(x)
    l_b_params = (k, 1 / n)
    r_b_params = (k + 1, 1 / n)
    cdfs = (gamma(*l_b_params), gamma(*r_b_params))
    return cbox_from_extredists(
        cdfs, shape="gamma", extre_bound_params=[l_b_params, r_b_params]
    )


def CBpoisson(x):
    n = len(x)
    k = sum(x)

    cdfs = (nbinom(k, 1 - 1 / (n + 1)), nbinom(k + 1, 1 - 1 / (n + 1)))
    return pbox_from_extredists(cdfs, shape="nbinom")


# * ---------------------exponential---------------------*#
# x[i] ~ exponential(parameter), x[i] is a nonnegative integer


def CBexponential_lambda(x):
    n = len(x)
    k = sum(x)
    conf_dist = gamma(n, scale=1 / k)
    return cbox_from_extredists(conf_dist, shape="gamma", extre_bound_params=(n, 1 / k))


def CBexponential(x):
    n = len(x)
    k = sum(x)

    def gammaexponential(shape, rate=1, scale=None):
        if scale is None:
            scale = 1 / rate
        rate = 1 / scale
        # expon(scale=gamma(a=shape, scale=1/rate))
        return scipy.stats.expon.rvs(
            scale=1 / scipy.stats.gamma.rvs(a=shape, scale=scale, size=Params.many),
            size=Params.many,
        )

    pseudo_s = gammaexponential(shape=n, rate=k)
    return pbox_from_pseudosamples(pseudo_s)


# * ---------------------normal---------------------*#


# x[i] ~ normal(mu, sigma)
def cboxNormalMu_base(x):
    """base function for precise sample x"""
    n = len(x)
    xm = np.mean(x)
    s = np.std(x)
    conf_dist = t(n - 1, loc=xm, scale=s / np.sqrt(n))  # conf_dist --> cd
    params = {"shape": "t", "loc": xm, "scale": (s / np.sqrt(n))}
    return conf_dist, params

    #! --- below is the nonparametric return style --- #!
    # x_support = rv.ppf(Params.p_values)
    # return x_support, params


@interval_measurements
def CBnormal_mu(x, style="analytical"):
    #! note the 'style' arg has no effect due to the decorator
    """
    args:
        x: (array-like) the sample data
        style: (str) the style of the output CDF, either 'analytical' or 'samples'
        size: (int) the discritisation size.
            meaning the no of ppf in analytical style and the no of MC samples in samples style

    return:
        CDF: (array-like) the CDF of the normal distribution
    """

    match style:
        case "analytical":
            # if isinstance(x, (list, np.ndarray)):
            #     x_sup = cboxNormalMu_base(x)
            #     return Cbox(left=x_sup, shape="t")

            # elif isinstance(x, Interval):
            #     x_sup_lo = cboxNormalMu_base(x.lo)
            #     x_sup_hi = cboxNormalMu_base(x.hi)
            #     return Cbox(left = x_sup_lo, right = x_sup_hi, shape="t")
            return cboxNormalMu_base(x)
        case "samples":
            n = len(x)

            def student(v):
                return scipy.stats.t.rvs(v, size=Params.many)

            # pop or sample std?
            return cbox_from_pseudosamples(
                np.mean(x) + np.std(x) * student(n - 1) / np.sqrt(n)
            )


def CBnormal_sigma(x):
    # TODO the analytical distribution equation
    def chisquared(v):
        return scipy.stats.chi2.rvs(v, size=Params.many)

    def inversechisquared(v):
        return 1 / chisquared(v)

    n = len(x)
    # pop or sample var?
    pseudo_s = np.sqrt(np.var(x) * (n - 1) * inversechisquared(n - 1))
    return cbox_from_pseudosamples(pseudo_s)


def CBnormal(x):
    n = len(x)

    # pop or sample std?
    def student(v):
        return scipy.stats.t.rvs(v, size=Params.many)

    return pbox_from_pseudosamples(
        np.mean(x) + np.std(x) * student(n - 1) * np.sqrt(1 + 1 / n)
    )


# * ---------------------lognormal---------------------*#

# x[i] ~ lognormal(mu, sigma), x[i] is a positive value whose logarithm is distributed as normal(mu, sigma)


def CBlognormal(x):
    n = len(x)
    return pbox_from_pseudosamples(
        np.exp(
            np.mean(np.log(x))
            + np.std(np.log(x)) * d.student(n - 1) * np.sqrt(1 + 1 / n)
        )
    )


def CBlognormal_mu(x):
    n = len(x)
    return cbox_from_pseudosamples(
        np.mean(np.log(x)) + np.std(np.log(x)) * d.student(n - 1) / np.sqrt(n)
    )


def CBlognormal_sigma(x):
    n = len(x)
    return cbox_from_pseudosamples(
        np.sqrt(np.var(np.log(x)) * (n - 1) * d.inversechisquared(n - 1))
    )


# * ---------------------uniform---------------------*#


# x[i] ~ uniform(midpoint, width)
# x[i] ~ uniform(minimum, maximum)


def CBuniform_midpoint(x):

    r = max(x) - min(x)
    w = r / d.beta(len(x) - 1, 2)
    m = (max(x) - w / 2) + (w - (max(x) - min(x))) * d.uniform(0, 1)
    return cbox_from_pseudosamples(m)


def CBuniform_width(x):
    r = max(x) - min(x)
    return cbox_from_pseudosamples(r / d.beta(len(x) - 1, 2))


def CBuniform_minimum(x):
    r = max(x) - min(x)
    w = r / d.beta(len(x) - 1, 2)
    m = (max(x) - w / 2) + (w - r) * d.uniform(0, 1)
    return cbox_from_pseudosamples(m - w / 2)


def CBuniform_maximum(x):
    r = max(x) - min(x)
    w = r / d.beta(len(x) - 1, 2)
    m = (max(x) - w / 2) + (w - r) * d.uniform(0, 1)
    return cbox_from_pseudosamples(m + w / 2)


# nextvalue


def CBuniform(x):
    r = max(x) - min(x)
    w = (r / d.beta(len(x) - 1, 2)) / 2
    m = (max(x) - w) + (2 * w - r) * d.uniform(0, 1)
    return pbox_from_pseudosamples(d.uniform(m - w, m + w))


# * ---------------------nonparametric---------------------*#

# x[i] ~ F, a continuous but unknown distribution
# N.B. the infinities don't plot, but they are there


def CBnonparametric(x):
    # TODO make below native to UN
    def env(x, y):
        return np.concatenate((x, y))

    def histogram(x):
        return x[(np.trunc(scipy.stats.uniform.rvs(size=2000) * len(x))).astype(int)]

    return env(
        histogram(np.concatenate((x, [-np.inf]))),
        histogram(np.concatenate((x, [np.inf]))),
    )


# x1[i] ~ normal(mu1, sigma1), x2[j] ~ normal(mu2, sigma2), x1 and x2 are independent


def CBnormal_meandifference(x1, x2):
    return CBnormal_mu(x2) - CBnormal_mu(x1)


# x[i] = Y + error[i],  error[j] ~ F,  F unknown,  Y fixed,  x[i] and error[j] are independent
def CBnonparametric_deconvolution(x, error):  # i.e., the c-box for Y

    def Get_Q(m_in, c_in, k=None):
        if k is None:
            k = np.arange((m_in * c_in + 1))

        def Q_size_GLBL(m):
            return 1 + m + m * (m + 1) / 2 + m * (m + 1) * (m + 2) * (3 * m + 1) / 24

        def Q_size_LoCL(m, c):
            return 1 + c + m * c * (c + 1) / 2

        def Grb_Q(m_in, c_in, Q_list):
            m = max(m_in, c_in)
            c = min(m_in, c_in)
            i_min = Q_size_GLBL(m - 1) + Q_size_LoCL(m, c - 1) + 1
            return Q_list[i_min : (i_min + m * c)]

        def AddingQ(m, Q_list):
            Q_list[Q_size_GLBL(m - 1) + 1] = 1
            for c in range(m):
                i_min = Q_size_GLBL(m - 1) + Q_size_LoCL(m, c) + 1
                Q1 = np.concatenate(
                    (Grb_Q(m - 1, c + 1, Q_list), np.repeat(0, (c + 1)))
                )
                Q2 = np.concatenate((np.repeat(0, m), Grb_Q(m, c, Q_list)))
                Q_list[i_min : (i_min + m * (c + 1))] = Q1 + Q2
            return Q_list[(Q_size_GLBL(m - 1) + 1) : Q_size_GLBL(m)]

        def Bld_Q(m_top):
            Q_out = np.repeat(0, Q_size_GLBL(m_top))
            Q_out[0] = 1
            for m in range(m_top):
                Q_out[(Q_size_GLBL(m) + 1) : (Q_size_GLBL(m + 1))] = AddingQ(
                    m + 1, Q_out
                )
            return Q_out

        # body of Get_Q
        m = max(m_in, c_in)
        c = min(m_in, c_in)
        return Grb_Q(m, c, Bld_Q(m))[k + 1]

    # body of CBnonparametric_deconvolution
    z = []
    for err in error:
        z = np.append(z, [x - err])
    z.sort()
    Q = Get_Q(len(x), len(error))
    w = Q / sum(Q)
    return env(mixture(z, w), mixture(np.append(z[1:], [np.inf]), w))


# * ---------------------helper modules---------------------*#


named_cbox = {
    "bernoulli": CBbernoulli_p,
    "binomial": CBbinomial_p,
    "exponential": CBexponential_lambda,
    "expon": CBexponential_lambda,
    # 'geometric': MMgeometric,
    "lognormal": {"mu": CBlognormal_mu, "sigma": CBlognormal_sigma},
    "norm": {"mu": CBnormal_mu, "sigma": CBnormal_sigma},
    "normal": {"mu": CBnormal_mu, "sigma": CBnormal_sigma},
    "gaussian": {"mu": CBnormal_mu, "sigma": CBnormal_sigma},
    "poisson": CBpoisson_lambda,
    "uniform": {
        "midpoint": CBuniform_midpoint,
        "width": CBuniform_width,
        "minimum": CBuniform_minimum,
        "maximum": CBuniform_maximum,
    },
}


named_nextvalue = {
    "bernoulli": CBbernoulli,
    "binomial": CBbinomial,
    "exponential": CBexponential,
    "expon": CBexponential,
    "lognormal": CBlognormal,
    "norm": CBnormal,
    "normal": CBnormal,
    "gaussian": CBnormal,
    "poisson": CBpoisson,
    "uniform": CBuniform,
}



from pyuncertainnumber import UncertainNumber as UN
from pyuncertainnumber.sensitivity.sobol_analysis import sobol_analysis


def test_sobol_analysis():
    # 1. Define uncertain inputs
    inputs = [
        UN(
            name="x1",
            essence="distribution",
            distribution_parameters=["uniform", (0, 1)],
        ),
        UN(
            name="x2",
            essence="distribution",
            distribution_parameters=["gaussian", (0.5, 0.1)],
        ),
    ]

    # 2. Define a model function (non-vectorized)
    def my_model(x):
        out = x[:, 0] + x[:, 1]
        return out

    # 3. Run the analysis
    results = sobol_analysis(
        inputs,
        my_model,
        plot_results=False,
        calc_second_order=True,
        print_to_console=True,
        N=512,
    )

    # 4. Print results for the first output
    print(results)


# def myFunctionWithTwoOutputs(x):
#     """Non-vectorized example function."""
#     input1, input2, input3, input4, input5 = x[0], x[1], x[2], x[3], x[4]
#     output1 = input1 + input2 + input3 + input4 + input5
#     output2 = input1 * input2 * input3 * input4 * input5
#     return output1#, output2

# x_inputs = [
#     UN(name='x0', essence='distribution', distribution_parameters=["gaussian", (1, 0.1)]),
#     UN(name='x1', essence='distribution', distribution_parameters=["gaussian", (2, 0.2)]),
#     UN(name='x2', essence='distribution', distribution_parameters=["gaussian", (3, 0.3)]),
#     UN(name='x3', essence='distribution', distribution_parameters=["gaussian", (4, 0.4)]),
#     UN(name='x4', essence='distribution', distribution_parameters=["gaussian", (5, 0.5)]),
# ]

# # Call the analysis function.
# results = sobol_analysis(
#     x_inputs,
#     myFunctionWithTwoOutputs,
#     output_names=['Sum'], #'Product'
#     N=2048
# )



import numpy
import scipy.stats as stats
import pickle
from pathlib import Path

def banana_data(n=100,d=2,seed:int=None, dist='normal'):
    def bake_banana_data(x:numpy.ndarray):
        y = numpy.asarray(x,dtype=float)
        Y1,Y2 = y[:,0]*4, y[:,0]**2+y[:,1]
        y[:,0], y[:,1] = Y1, Y2
        return y
    if dist == 'normal': dist = stats.norm(loc=0,scale=2)
    else: dist = stats.norm(loc=0,scale=2) # todo: add more distributions
    if seed is not None: x_input = dist.rvs(size=(n,d),seed=seed)  # x_input = 2*numpy.random.randn(N,2,)
    else: x_input = x_input = dist.rvs(size=(n,d))
    return bake_banana_data(x_input)

def pickle_dump(x:numpy.ndarray,filename:str=None,ext='.pickle',ordinal=True):
    dir = './data/'
    Path(dir).mkdir(parents=True,exist_ok=True)
    if filename is None: 
        if ordinal: filename = 'banana_'+str(get_next_integer(dir))
        else: filename = 'banana_'+str(numpy.random.randint(10_000))
    pickle.dump(x,open(dir+filename+ext,'wb'))
    pass

def get_next_integer(subdirectory: str):
    def next_integer(x):
        if len(x)==0: return 1
        else: return max(x)+1
    pathlist = Path(subdirectory).glob('*.pickle')
    ordinal = []
    for file in pathlist:
        index = pull_first_integer(str(file).split('_'))
        ordinal.append(int(index))
    return next_integer(ordinal)

def pull_first_integer(l:list):
    for s in l:
        s_=s.split('.')
        try: return int(s_[0])
        except: pass


def pickle_load(filename:str,ext='.pickle'):
    dir = './data/'
    Path(dir).mkdir(parents=True,exist_ok=True)
    return pickle.load(open(dir+filename+ext,'rb'))

def banana_model(x:numpy.ndarray):
    x = numpy.asarray(x,dtype=float)
    y = numpy.empty(x.shape,dtype=float)
    Y1,Y2 = x[:,0]*4, x[:,0]**2+x[:,1]
    y[:,0], y[:,1] = Y1, Y2
    return y


if __name__=='__main__':
    x = banana_data()
    # pickle_dump(x)


from .calibration import *
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
import numpy as np
from sklearn.preprocessing import StandardScaler
from sklearn.neighbors import NearestNeighbors, KernelDensity


class KNNCalibrator(Calibrator):
    """Unified kNN-based calibrator for black-box models or precomputed simulations.

    Args:
        knn (int): Number of neighbors per observed row. Default: 100.
        a_tol (float): Tolerance for matching simulated :math:`\\xi` to a requested :math:`\\xi^*`
            (when reusing). A simulation is kept if :math:`\\|\\xi_{\\text{sim}} - \\xi^*\\|_\\infty < a_{\\text{tol}}`.
            Default: 0.05.
        evaluate_model (bool): If True, call the black-box model for each :math:`\\xi` in ``xi_list``
            on a shared :math:`\\theta` grid. If False, reuse ``simulated_data`` (requires y/theta/xi arrays).
            Default: False.
        random_state (int): Seed for reproducibility (affects theta_sampler and resampling). Default: 42.

    Note:
        **Setup (unified approach)**:

        - If ``evaluate_model=False`` and ``simulated_data`` is provided:

          * Reuse pre-computed simulations
          * Build a per-design kNN index by filtering rows with :math:`\\|\\xi - \\xi^*\\|_\\infty < a_{\\text{tol}}`
            for each :math:`\\xi^*` in ``xi_list``

        - If ``evaluate_model=True``:

          * Simulate :math:`y = \\text{model}(\\theta, \\xi)` for each :math:`\\xi` in ``xi_list``
          * Use a shared :math:`\\theta` grid drawn once from ``theta_sampler(n_samples)``
          * Build per-design kNN indices on this shared grid

        **Calibration workflow (single/multi-design)**:

        For each observation pair :math:`(y_{\\text{obs}}, \\xi)`:

        1. Standardize :math:`y_{\\text{obs}}` using the per-design scaler
        2. Find k nearest neighbors in y-space
        3. Map neighbor indices to :math:`\\theta` values for that design
        4. Stack :math:`\\theta` samples across all observations/designs (or apply voting/intersection)

    .. figure:: /_static/knn_illustration.png
        :alt: knn_illustration
        :align: center
        :width: 50%

        KNN calibration illustration.
    """

    def __init__(
        self, knn: int = 100, a_tol: float = 0.05, evaluate_model: bool = False
    ):

        super().__init__()
        self.knn = int(knn)
        self.a_tol = float(a_tol)
        self.evaluate_model = bool(evaluate_model)
        self.random_state = 42

        # Internal state
        self._theta_grid: Optional[np.ndarray] = (
            None  # shared grid if evaluate_model=True, else unused
        )
        self._theta_by_xi: Dict[Tuple[float, ...], np.ndarray] = (
            {}
        )  # per-design θ (may be shared ref)
        self._y_by_xi: Dict[Tuple[float, ...], np.ndarray] = {}  # per-design y
        self._scaler_by_xi: Dict[Tuple[float, ...], StandardScaler] = {}
        self._neigh_by_xi: Dict[Tuple[float, ...], NearestNeighbors] = {}
        self._grid_idx_by_xi: Dict[Tuple[float, ...], np.ndarray] = {}
        self._posterior: Optional[Dict[str, Any]] = None

        # Keep original sims if reusing
        self._sim_y: Optional[np.ndarray] = None
        self._sim_theta: Optional[np.ndarray] = None
        self._sim_xi: Optional[np.ndarray] = None

    # ---------- utilities ----------
    @staticmethod
    def _key_from_xi(xi) -> Tuple[float, ...]:
        """Stable tuple key for a scalar/vector design ξ."""
        return tuple(np.atleast_1d(np.asarray(xi, float)).ravel())

    # ---------- setup ----------
    def setup(
        self,
        model: Optional[
            Callable[[np.ndarray, Union[float, np.ndarray]], np.ndarray]
        ] = None,
        theta_sampler: Optional[Callable[[int], np.ndarray]] = None,
        simulated_data: Optional[Dict[str, np.ndarray]] = None,
        xi_list: Optional[List[Union[float, np.ndarray]]] = None,
        n_samples: int = 10000,
    ):
        """Prepare per-design kNN structures by either reusing simulated_data or simulating for each design.

        Args:
            model (callable, optional): Black-box simulator with signature ``model(theta, xi) -> y``
                (vectorized over theta).
            theta_sampler (callable, optional): Sampler for :math:`\\theta`; required when ``evaluate_model=True``.
            simulated_data (dict, optional): Dict with keys {"y": (n, dy), "theta": (n, dθ), "xi": (n, dξ)}
                when reusing sims.
            xi_list (list, optional): List of designs; each item can be scalar or array-like.
                If None, defaults to [0.0].
            n_samples (int): Number of :math:`\\theta` samples to draw when ``evaluate_model=True``. Default: 10000.
        """
        xi_list = [0.0] if not xi_list else xi_list

        # Reset state
        self._theta_grid = None
        self._theta_by_xi.clear()
        self._y_by_xi.clear()
        self._scaler_by_xi.clear()
        self._neigh_by_xi.clear()
        self._posterior = None

        if not self.evaluate_model:
            # ---- Reuse provided simulations; filter per design ----
            if simulated_data is None:
                raise ValueError(
                    "evaluate_model=False requires `simulated_data` with keys 'y','theta','xi'."
                )

            self._sim_y = np.asarray(simulated_data["y"], float)
            self._sim_theta = np.asarray(simulated_data["theta"], float)
            self._sim_xi = np.asarray(simulated_data.get("xi", None), float)
            if self._sim_xi is None:
                raise ValueError(
                    "`simulated_data` must include 'xi' to filter per design."
                )

            for xi in xi_list:
                key = self._key_from_xi(xi)
                mask = np.all(
                    np.abs(self._sim_xi - np.atleast_1d(xi)) < self.a_tol, axis=1
                )
                y_xi = self._sim_y[mask]
                theta_xi = self._sim_theta[mask]
                if y_xi.size == 0:
                    raise ValueError(
                        f"No simulations matched design {xi} within tolerance a_tol={self.a_tol}."
                    )
                # drop NaNs rows in y
                ok = ~np.isnan(y_xi).any(axis=1)
                y_xi, theta_xi = y_xi[ok], theta_xi[ok]
                if y_xi.size == 0:
                    raise ValueError(f"All simulations at design {xi} had NaNs in y.")
                # build scaler & kNN
                sc = StandardScaler().fit(y_xi)
                neigh = NearestNeighbors(n_neighbors=self.knn).fit(sc.transform(y_xi))
                # store
                self._theta_by_xi[key] = theta_xi
                self._y_by_xi[key] = y_xi
                self._scaler_by_xi[key] = sc
                self._neigh_by_xi[key] = neigh

        else:
            # ---- Evaluate model per design on a shared θ grid ----
            if model is None or theta_sampler is None:
                raise ValueError(
                    "evaluate_model=True requires `model` and `theta_sampler`."
                )
            self._theta_grid = np.asarray(theta_sampler(int(n_samples)), float)
            if self._theta_grid.ndim != 2:
                raise ValueError(
                    "theta_sampler must return a 2D array (n_samples, dθ)."
                )

            for xi in xi_list:
                key = self._key_from_xi(xi)
                # vectorized model call over θ
                y_xi = np.asarray(model(self._theta_grid, xi), float)
                if y_xi.ndim == 1:
                    y_xi = y_xi[:, None]
                if y_xi.shape[0] != self._theta_grid.shape[0]:
                    raise ValueError("Model must return one row of y per θ sample.")
                # drop NaNs rows in y (and corresponding θ rows)
                ok = ~np.isnan(y_xi).any(axis=1)
                y_xi = y_xi[ok]
                theta_xi = self._theta_grid[ok]
                grid_idx = np.where(ok)[0]
                self._grid_idx_by_xi[key] = (
                    grid_idx  # maps local row j -> global grid index grid_idx[j]
                )

                if y_xi.size == 0:
                    raise ValueError(f"All simulations at design {xi} had NaNs in y.")
                # build scaler & kNN
                sc = StandardScaler().fit(y_xi)
                neigh = NearestNeighbors(n_neighbors=self.knn).fit(sc.transform(y_xi))
                # store
                self._theta_by_xi[key] = theta_xi
                self._y_by_xi[key] = y_xi
                self._scaler_by_xi[key] = sc
                self._neigh_by_xi[key] = neigh

        self.is_ready = True

    # ---------- nearest ----------
    def nearest(
        self,
        y: Union[np.ndarray, List[float]],
        xi: Union[float, np.ndarray],
        k: Optional[int] = None,
        return_dist: bool = False,
    ):
        """Return k nearest neighbors for y at design xi.

        Args:
            y (array-like): Query outputs, shape (m, d_y) or (d_y,).
            xi (scalar or array-like): Design key to select the per-design index.
            k (int, optional): Number of neighbors; defaults to ``self.knn``.
            return_dist (bool): If True, also return distances and raw indices. Default: False.

        Returns:
            theta_neighbors (ndarray): Shape (m*k, dθ) stacked :math:`\\theta` for all query rows.
            distances (ndarray, optional): Returned if ``return_dist=True``.
            indices (ndarray, optional): Returned if ``return_dist=True``.
        """
        if not self.is_ready:
            raise RuntimeError("Call setup() before nearest().")
        key = self._key_from_xi(xi)
        if key not in self._neigh_by_xi:
            raise KeyError(
                f"Design {xi} not in index. Known: {list(self._neigh_by_xi.keys())}"
            )
        y = np.atleast_2d(np.asarray(y, float))
        sc = self._scaler_by_xi[key]
        neigh = self._neigh_by_xi[key]
        k_eff = int(k or self.knn)
        d, idx = neigh.kneighbors(sc.transform(y), n_neighbors=k_eff)
        theta = self._theta_by_xi[key]
        theta_neighbors = np.vstack([theta[i] for i in idx])
        if return_dist:
            return theta_neighbors, d, idx
        return theta_neighbors

    # ---------- calibration ----------
    def calibrate(
        self,
        observations,
        resample_n: int | None = None,
        combine: str = "stack",
        combine_params: dict | None = None,
    ):
        """Run kNN calibration and aggregate posterior θ across neighbor-hit blocks.

        Args:
            observations: Observed simulator or model outputs to calibrate against.

            resample_n (int | None): If set, resample posterior θ samples to this size.
                If `None`, return all aggregated θ without resampling.

            combine (str): Aggregation mode. One of:

                - **'stack'**: concatenate all kNN θ; optional de-duplication.

                - **'intersect'**: retain θ hit at least `min_count` times across neighbor blocks.

            combine_params (dict | None): Optional parameters controlling aggregation and KDE weighting.

                Supported keys:

                - **dedup** (bool): Default `False`. Remove duplicate θ (only for 'stack').

                - **theta_match_tol** (float): Default `1e-9`. Tolerance or rounding quantum for comparing/merging θ values.

                - **min_count** (int | None): Minimum occurrences for 'intersect'. Default is `max(1, ceil(0.5 * total_blocks))`, meaning θ must appear in about half of neighbor lists.

                - **use_kde** (bool): Default `False`. If `True`, fit KDE on aggregated θ to compute log-scores and normalized weights.

                - **kde_bandwidth** (float | None): Bandwidth for KDE. If `None` (default), use Scott's rule.

        tip:
            Two aggregation modes are supported:

            - **stack**: Concatenate all kNN θ into a single array. Supports optional de-duplication of nearly identical θ values.

            - **intersect**: Keep θ values that occur in at least `min_count` neighbor blocks across all observations/design points (default ≈ half of all blocks).

            Optional density weighting via KDE can be applied after aggregation to compute normalized posterior weights.


        Returns:
            dict: A dictionary with keys:
                - **'mode'** (str): Always `'knn'`.
                - **'theta'** (ndarray): Posterior samples of shape `(N, dθ)`;
                  resampled if `resample_n` is provided.
                - **'weights'** (ndarray | None): `None` for stack/intersect,
                  or a length-`N` array of KDE weights if `use_kde=True`.
                - **'meta'** (dict): Aggregation info; may include KDE bandwidth
                  if density weighting is used.
        """

        if not self.is_ready:
            raise RuntimeError("Call setup() before calibrate().")

        combine_params = combine_params or {}
        dedup = bool(combine_params.get("dedup", False))
        tol = float(combine_params.get("theta_match_tol", 1e-9))
        use_kde = bool(combine_params.get("use_kde", True))
        kde_bw = combine_params.get("kde_bandwidth", 0.1)

        # ---------------- Collect θ-neighbors for every (y, ξ) ----------------
        theta_hits = (
            []
        )  # list of (n_i*k, dθ) blocks, one block per y-row (across all designs)
        for y_obs, xi in observations:
            key = self._key_from_xi(xi)
            if key not in self._neigh_by_xi:
                raise KeyError(
                    f"Design {xi} not in index. Known: {list(self._neigh_by_xi.keys())}"
                )
            yo = np.atleast_2d(np.asarray(y_obs, float))
            yo = yo[~np.isnan(yo).any(axis=1)]
            if yo.size == 0:
                continue
            sc, neigh = self._scaler_by_xi[key], self._neigh_by_xi[key]
            d, idx = neigh.kneighbors(
                sc.transform(yo), n_neighbors=self.knn, return_distance=True
            )
            # gather θ for this design
            th = self._theta_by_xi[key]
            # flatten all rows’ neighbors for this block
            theta_block = np.vstack([th[i] for i in idx])  # (n_rows*k, dθ)
            theta_hits.append(theta_block)

        if len(theta_hits) == 0:
            raise ValueError("No valid observations after NaN filtering.")

        # ---------------- Aggregation strategies ----------------
        if combine == "stack":
            theta_all = np.vstack(theta_hits)  # (sum n_i*k, dθ)

            if dedup:
                uniq, _ = self._round_rows(theta_all, tol)
                theta_out = uniq
            else:
                theta_out = theta_all

            # Optional KDE scoring on returned support
            weights = None
            meta = {"combine": "stack", "dedup": dedup, "theta_match_tol": tol}
            if use_kde and theta_out.shape[0] > 0:
                logp, w = self._kde_logweights(theta_out, bw=kde_bw)
                weights = w
                meta.update({"use_kde": True, "kde_bandwidth": kde_bw})

            # Optional resampling
            if resample_n and theta_out.shape[0] > 0:
                rng = np.random.default_rng(self.random_state)
                if weights is None:
                    take = rng.choice(
                        theta_out.shape[0], size=int(resample_n), replace=True
                    )
                else:
                    take = rng.choice(
                        theta_out.shape[0],
                        size=int(resample_n),
                        replace=True,
                        p=weights,
                    )
                theta_out = theta_out[take]

            self._posterior = {
                "mode": "knn",
                "theta": theta_out,
                "weights": weights,
                "meta": meta,
            }
            return self._posterior

        elif combine == "intersect":
            # Build one big stack and count approximate matches
            big = np.vstack(theta_hits)  # (M, dθ)
            uniq, counts = self._round_rows(big, tol)

            # total neighbor lists (one per row across all designs)
            total_blocks = sum(b.shape[0] // self.knn for b in theta_hits)

            # Strictness knobs
            min_frac = float(
                combine_params.get("min_frac", 0.8)
            )  # keep θ seen in ≥80% of lists
            min_count = combine_params.get("min_count", None)
            if min_count is None:
                min_count = max(1, int(np.ceil(min_frac * total_blocks)))

            # Filter by frequency
            keep = counts >= int(min_count)
            theta_out = uniq[keep]
            counts_sel = counts[keep].astype(float)

            # If nothing passed, fall back to TOP-FRACTION
            if theta_out.shape[0] == 0:
                top_frac = float(
                    combine_params.get("top_frac", 0.1)
                )  # keep top 10% by frequency
                k = max(1, int(np.ceil(top_frac * len(counts))))
                top_idx = np.argsort(counts)[::-1][:k]
                theta_out = uniq[top_idx]
                counts_sel = counts[top_idx].astype(float)
                meta = {
                    "combine": "intersect",
                    "theta_match_tol": tol,
                    "min_count": min_count,
                    "min_frac": min_frac,
                    "fallback": f"top-{top_frac:.2f}",
                }
            else:
                meta = {
                    "combine": "intersect",
                    "theta_match_tol": tol,
                    "min_count": min_count,
                    "min_frac": min_frac,
                }

            # Frequency-based weights (sharpen with gamma)
            weights = None
            if theta_out.shape[0] > 0:
                gamma = float(
                    combine_params.get("gamma", 1.0)
                )  # 1.0=no sharpen, 2.0=stricter
                w_counts = counts_sel ** max(gamma, 1e-12)

                # Optional: KDE blending for smoother density
                if bool(combine_params.get("use_kde", False)):
                    kde_bw = combine_params.get("kde_bandwidth", 0.1)
                    _, w_kde = self._kde_logweights(theta_out, bw=kde_bw)
                    beta = float(
                        combine_params.get("beta", 1.0)
                    )  # blend exponent for KDE
                    w = w_counts * (w_kde**beta)
                    w = np.asarray(w, float)
                    w = w / (w.sum() if w.sum() > 0 else len(w))
                    weights = w
                    meta.update(
                        {
                            "use_kde": True,
                            "kde_bandwidth": kde_bw,
                            "gamma": gamma,
                            "beta": beta,
                        }
                    )
                else:
                    w = w_counts / (
                        w_counts.sum() if w_counts.sum() > 0 else len(w_counts)
                    )
                    weights = w
                    meta.update({"gamma": gamma})

            # Optional resampling
            if resample_n and theta_out.shape[0] > 0:
                rng = np.random.default_rng(self.random_state)
                if weights is None:
                    take = rng.choice(
                        theta_out.shape[0], size=int(resample_n), replace=True
                    )
                else:
                    take = rng.choice(
                        theta_out.shape[0],
                        size=int(resample_n),
                        replace=True,
                        p=weights,
                    )
                theta_out = theta_out[take]

            self._posterior = {
                "mode": "knn",
                "theta": theta_out,
                "weights": weights,
                "meta": meta,
            }
            return self._posterior

        else:
            raise ValueError("`combine` must be 'stack' or 'intersect'.")

    def _round_rows(self, A: np.ndarray, tol: float) -> tuple[np.ndarray, np.ndarray]:
        """Round rows of A to multiples of tol and return (unique_rows, counts).

        Args:
            A (ndarray): Input array to process.
            tol (float): Tolerance for rounding. If tol <= 0, exact matching is used.

        Returns:
            tuple: (unique_rows, counts) where unique_rows are the deduplicated rows
                and counts are the occurrence counts.
        """
        import numpy as _np

        A = _np.asarray(A, float)
        if A.size == 0:
            return A.copy(), _np.array([], dtype=int)
        if tol <= 0:
            uniq, idx, counts = _np.unique(
                A, axis=0, return_index=True, return_counts=True
            )
            order = _np.sort(idx)
            uniq = A[order]
            counts = counts[_np.argsort(idx)]
            return uniq, counts
        R = _np.round(A / tol) * tol
        uniq, idx, counts = _np.unique(R, axis=0, return_index=True, return_counts=True)
        order = _np.sort(idx)
        uniq = R[order]
        counts = counts[_np.argsort(idx)]
        return uniq, counts

    def _kde_logweights(self, X, bw=0.5, n_max_exact=5000):
        """Compute KDE-based log-weights for posterior samples X.

        Args:
            X (ndarray): Posterior samples, shape (n, d).
            bw (float): Bandwidth for Gaussian kernel. Default: 0.5.
            n_max_exact (int): Max n for exact pairwise KDE. Above this, fall back to
                sklearn.KernelDensity. Default: 5000.

        Returns:
            tuple:
                - **logp** (ndarray): Log-density values at X, shape (n,).
                - **w** (ndarray): Normalized weights, shape (n,).
        """
        n, d = X.shape
        if n <= n_max_exact:
            # ---- Exact method (safe for small n) ----
            h2 = float(bw) ** 2
            d2 = np.sum((X[:, None, :] - X[None, :, :]) ** 2, axis=2)  # (n,n)
            K = np.exp(-0.5 * d2 / (h2 + 1e-18))
            sK = K.sum(axis=1) + 1e-300
            logp = np.log(sK)
            w = sK / sK.sum()
        else:
            # ---- Scalable method (sklearn KD-tree backend) ----
            kde = KernelDensity(kernel="gaussian", bandwidth=bw)
            kde.fit(X)
            logp = kde.score_samples(X)  # log density at each sample
            w = np.exp(logp - logp.max())
            w /= w.sum()

        return logp, w

    # ---------- posterior ----------
    def get_posterior(self) -> Any:
        """Return the last computed posterior dict; raises if calibrate() hasn't been called."""
        if self._posterior is None:
            raise RuntimeError("No posterior available. Run calibrate() first.")
        return self._posterior


def estimate_p_theta_knn(
    observed_data, simulated_data, xi_star, knn: int = 20, a_tol: float = 0.05
):
    """Estimate the posterior distribution p(θ) using k-Nearest Neighbors (kNN) on a simulation archive.

    This method restricts the simulation archive to runs at (or near) the target design :math:`\\xi^*`, 
    then fits a kNN model in output (y) space. For each observed output y_obs, it retrieves the 
    k-nearest simulated outputs and returns the corresponding :math:`\\theta` values as approximate 
    posterior samples.

    Args:
        observed_data (ndarray): Array of observed outputs y_obs, shape (n_obs, d_y).
            Must match the dimensionality of simulated outputs.
        simulated_data (list): List of arrays [y, θ, ξ], containing:
            
            - **y** (ndarray): Simulation output, shape (n, d_y), e.g., transformed y with only KPIs
            - **θ** (ndarray): Parameters and variables to be calibrated, shape (n, d_theta)
            - **ξ** (ndarray): Conditioning controllable factors, shape (n, d_xi), e.g., design parameters
            
        knn (int): Number of nearest neighbors to query per observed sample. Default: 20.
        xi_star (scalar or array-like): Target design :math:`\\xi^*` at which the posterior is estimated.
        a_tol (float): Tolerance for matching simulations to :math:`\\xi^*`. Default: 0.05.
            A simulation is kept if :math:`\\|\\xi_{\\text{sim}} - \\xi^*\\|_\\infty < a_{\\text{tol}}`.

    Returns:
        ndarray: :math:`\\theta` samples from the posterior, stacked across all observed y.
            Shape: (n_obs × knn, d_theta).

    Raises:
        ValueError: If filtering leaves no simulations at :math:`\\xi^*`.
        RuntimeError: If kNN search fails due to inconsistent dimensions.

    Note:
        - Scaling of outputs y is performed internally via StandardScaler for robustness 
          against different KPI magnitudes.
        - The parameter ``knn`` acts as a smoothing parameter: higher values broaden the 
          posterior but reduce sharpness.
        - The choice of ``a_tol`` trades off strict design conditioning vs. sample size. 
          Too small → few matches; too large → weaker conditioning.

    Example:
        >>> import numpy as np
        >>> from sklearn.preprocessing import StandardScaler
        >>> from sklearn.neighbors import NearestNeighbors
        >>> # Fake simulator archive
        >>> theta_sim = np.random.uniform(-5, 5, size=(5000, 2))
        >>> xi_sim = np.zeros((5000, 1))
        >>> y_sim = np.sum(theta_sim**2, axis=1, keepdims=True) \
        ...         + 0.1*np.random.randn(5000, 1)
        >>> simulated_data = [y_sim, theta_sim, xi_sim]
        >>> # Observed data
        >>> theta_true = np.array([1.5, -2.0])
        >>> y_obs = np.sum(theta_true**2) + 0.1*np.random.randn(1)
        >>> # Estimate posterior
        >>> theta_post = estimate_p_theta_knn(
        ...     observed_data=np.array([[y_obs]]),
        ...     simulated_data=simulated_data,
        ...     knn=50,
        ...     xi_star=0.0
        ... )
        >>> theta_post.shape
        (50, 2)
        >>> theta_post.mean(axis=0)
        array([ 1.4, -2.1])  # close to true [1.5, -2.0]
    """

    # Step 1: Filter simulated datasets based on ξ = ξ*
    xi_idx = np.all(np.abs(simulated_data[2] - xi_star) < a_tol, axis=1)
    simulated_data_xi = [s[xi_idx] for s in simulated_data]

    # Step 2: fit a kNN on the (filtered) space of y. Normalize observations
    scaler = StandardScaler()
    if np.any(np.isnan(simulated_data_xi[0])):
        simulated_data_xi[0] = simulated_data_xi[0][
            ~np.isnan(simulated_data_xi[0]).any(axis=1)
        ]

    scaler.fit(simulated_data_xi[0])
    neigh = NearestNeighbors(n_neighbors=knn)
    neigh.fit(scaler.transform(simulated_data_xi[0]))

    # Step 3: retrieve the kNN for each observed y_i  ...... check if there are nan values in the observed datasets
    if np.any(np.isnan(observed_data)):
        observed_data = observed_data[~np.isnan(observed_data).any(axis=1)]
    dist, knn_idx = neigh.kneighbors(scaler.transform(observed_data))
    theta_set = np.vstack([simulated_data_xi[1][idx] for idx in knn_idx])
    return theta_set



from .number import Interval
from .number import lo, hi, mid, rad, width
from .methods import (
    contain,
    intersect,
    straddle_zero,
    split_interval,
    subintervalise,
    intervalise,
    sizeit,
)
from .random import uniform_endpoints



import numpy as np
import pytest

from pyuncertainnumber import pba

""" Numpy support tests"""

# *  --------------------- for pbox


@pytest.fixture
def example_staircase():
    return pba.normal(5, 1)


# unary funcs: {sin, cos, tanh, exp, log, sqrt}
def test_sin_matches_method(example_staircase):
    d = example_staircase
    assert np.sin(d) == d.sin()


def test_cos_matches_method(example_staircase):
    d = example_staircase
    assert np.cos(d) == d.cos()


def test_tanh_matches_method(example_staircase):
    d = example_staircase
    assert np.tanh(d) == d.tanh()


def test_exp_matches_method(example_staircase):
    d = example_staircase
    assert np.exp(d) == d.exp()


def test_log_matches_method(example_staircase):
    d = example_staircase
    assert np.log(d) == d.log()


def test_sqrt_matches_method(example_staircase):
    d = example_staircase
    assert np.sqrt(d) == d.sqrt()


# *  --------------------- for Interval


# *  --------------------- for Distribution


# *  --------------------- for UN



# * --------------------- dependency context manager ---------------------*#

from contextvars import ContextVar
from contextlib import contextmanager

# Global context variable for dependency type
_current_dependency = ContextVar("current_dependency", default="f")


@contextmanager
def dependency(dep_type: str):
    """Context manager to temporarily change arithmetic dependency ('f', 'p', 'o', or 'i')"""
    token = _current_dependency.set(dep_type)
    try:
        yield
    finally:
        _current_dependency.reset(token)


def get_current_dependency():
    return _current_dependency.get()



from functools import wraps


class NominalValueMixin:
    @property
    def nominal_value(self):
        return self._compute_nominal_value()

    def _compute_nominal_value(self):
        raise NotImplementedError


# --- mixin that forwards ops to Pbox after converting self/other ---
class _PboxOpsMixin:
    """
    Mixin for classes that want to reuse Pbox's arithmetic by
    converting themselves (and the other operand) into Pbox first.
    Requires a _to_pbox(self) -> Pbox method on the subclass.
    """

    # if you also want comparisons, add them similarly
    _BIN_OPS = [
        "__add__",
        "__sub__",
        "__mul__",
        "__truediv__",
        "__floordiv__",
        "__mod__",
        "__pow__",
        "__matmul__",
        "__and__",
        "__or__",
        "__xor__",
        "__lshift__",
        "__rshift__",
    ]
    _REFL = {  # reflected names
        "__add__": "__radd__",
        "__sub__": "__rsub__",
        "__mul__": "__rmul__",
        "__truediv__": "__rtruediv__",
        "__floordiv__": "__rfloordiv__",
        "__mod__": "__rmod__",
        "__pow__": "__rpow__",
        "__matmul__": "__rmatmul__",
        "__and__": "__rand__",
        "__or__": "__ror__",
        "__xor__": "__rxor__",
        "__lshift__": "__rlshift__",
        "__rshift__": "__rrshift__",
    }
    _UNARY_OPS = ["__neg__", "__pos__", "__abs__", "__invert__"]

    @staticmethod
    def _coerce_to_pbox(x):
        from .pbox_abc import Pbox

        if isinstance(x, Pbox):
            return x
        # If the other operand is a DSS (or another _PboxOpsMixin), convert it too
        if isinstance(x, _PboxOpsMixin):
            return x._to_pbox()
        return x  # let Pbox handle plain numbers, numpy arrays, etc.

    # install methods on the class at definition time
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        from .pbox_abc import Pbox

        # binary ops
        for name in cls._BIN_OPS:
            refl = cls._REFL[name]

            def make_bin(nm):
                @wraps(getattr(Pbox, nm, lambda *a, **k: None))
                def _op(self, other):
                    a = self._to_pbox()
                    b = cls._coerce_to_pbox(other)
                    meth = getattr(a, nm, None)
                    if meth is None:
                        return NotImplemented
                    return meth(b)

                return _op

            def make_rbin(rnm):
                @wraps(getattr(Pbox, rnm, lambda *a, **k: None))
                def _rop(self, other):
                    a = cls._coerce_to_pbox(other)
                    b = self._to_pbox()
                    meth = getattr(b, rnm, None)
                    if meth is None:
                        return NotImplemented
                    return meth(a)

                return _rop

            setattr(cls, name, make_bin(name))
            setattr(cls, refl, make_rbin(refl))

        # unary ops
        for name in cls._UNARY_OPS:

            def make_unary(nm):
                @wraps(getattr(Pbox, nm, lambda *a, **k: None))
                def _u(self):
                    return getattr(self._to_pbox(), nm)()

                return _u

            setattr(cls, name, make_unary(name))

    # Optional: delegate *other* methods/attributes to the Pbox view
    def __getattr__(self, name):
        # Called only if normal lookup fails, so avoid shadowing real attrs
        return getattr(self._to_pbox(), name)
